"""Functions related to scattering matrix computation for the FMM algorithm.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import dataclasses
import functools
from typing import Sequence, Tuple

import jax
import jax.numpy as jnp
from jax import tree_util

from fmmax import fmm, utils


@dataclasses.dataclass
class ScatteringMatrix:
    """Stores the scattering matrix for a stack of layers.

    The first layer in a stack is the "start" layer, and the last layer in the
    stack is the "end" layer.

    The scattering matrix relates the forward-going and backward-going waves
    on the two sides of a layer stack, which are labeled `a` and `b` respectively.

    Note that forward going fields are defined at the *start* of a layer while
    backward-going fields are defined at the *end* of a layer, as depicted below.
    This is discussed near equation 4.1 in [1999 Whittaker].

                |             |           |         |           |
                |   layer 0   |  layer 1  |   ...   |  layer N  |
                | start layer |           |         | end layer |
                |             |           |         |           |
                 -> a_0                              -> a_N
                        b_0 <-                            b_N <-

    Following the convention of [1999 Whittaker], the terms a_N and b_0 are
    obtained from,

                    a_N = s11 @ a_0 + s12 @ b_N
                    b_0 = s21 @ a_0 + s22 @ b_N

    Besides the actual scattering matrix element, the `ScatteringMatrix` stores
    information about the start and end layers, which are needed to extend the
    scattering matrix to include more layers.

    Attributes:
        s11: Relates forward-going fields at start to forward-going fields at end.
        s12: Relates backward-going fields at end to forward-going fields at end.
        s21: Relates forward-going fields at start to backward-going fields at start.
        s22: Relates backward-going fields at end to backward-going fields at start.
        start_layer_solve_result: The eigensolve result for the start layer.
        start_layer_thickness: The start layer thickness.
        end_layer_solve_result: The eigensolve result for the end layer.
        end_layer_thickness: The end layer thickness.
    """

    s11: jnp.ndarray
    s12: jnp.ndarray
    s21: jnp.ndarray
    s22: jnp.ndarray

    start_layer_solve_result: fmm.LayerSolveResult
    start_layer_thickness: jnp.ndarray

    end_layer_solve_result: fmm.LayerSolveResult
    end_layer_thickness: jnp.ndarray


def stack_s_matrix(
    layer_solve_results: Sequence[fmm.LayerSolveResult],
    layer_thicknesses: Sequence[jnp.ndarray],
    force_x64_solve: bool = False,
) -> ScatteringMatrix:
    """Computes the s-matrix for a stack of layers.

    If only a single layer is provided, the scattering matrix is just the
    identity matrix, and start and end layer data is for the same layer.

    Args:
        layer_solve_results: The eigensolve results for layers in the stack.
        layer_thicknesses: The thicknesses for layers in the stack.
        force_x64_solve: If `True`, matrix solves will be done with 64 bit precision.

    Returns:
        The `ScatteringMatrix`.
    """
    return _stack_s_matrices(
        layer_solve_results, layer_thicknesses, force_x64_solve=force_x64_solve
    )[-1]


def stack_s_matrices_interior(
    layer_solve_results: Sequence[fmm.LayerSolveResult],
    layer_thicknesses: Sequence[jnp.ndarray],
    force_x64_solve: bool = False,
) -> Tuple[Tuple[ScatteringMatrix, ScatteringMatrix], ...]:
    """Computes scattering matrices before and after each layer in the stack.

    Specifically, for each layer `i` two `ScatteringMatrix` are returned. The
    first relates fields in the substack `0, ..., i`, while the second relates
    the fields in the substack `i, ..., N`, where `N` is the maximum layer
    index. These two scattering matrices are denoted as the corresponding
    to the "before" substack and the "after" substack.

    Args:
        layer_solve_results: The eigensolve results for layers in the stack.
        layer_thicknesses: The thicknesses for layers in the stack.
        force_x64_solve: If `True`, matrix solves will be done with 64 bit precision.

    Returns:
        The tuple of `(scattering_matrix_before, scattering_matrix_after)`.
    """
    before = _stack_s_matrices(
        layer_solve_results,
        layer_thicknesses,
        force_x64_solve=force_x64_solve,
    )

    # Compute the scattering matrix for the substack "after" each layer. We do
    # this by computing the scattering matrix for the substack "before" each layer
    # in the reversed stack. Then, reverse each resulting scattering matrix.
    reverse = _stack_s_matrices(
        layer_solve_results[::-1],
        layer_thicknesses[::-1],
        force_x64_solve=force_x64_solve,
    )
    after = tuple(
        [
            ScatteringMatrix(
                s11=s.s22,
                s12=s.s21,
                s21=s.s12,
                s22=s.s11,
                start_layer_solve_result=s.end_layer_solve_result,
                start_layer_thickness=s.end_layer_thickness,
                end_layer_solve_result=s.start_layer_solve_result,
                end_layer_thickness=s.start_layer_thickness,
            )
            for s in reverse[::-1]
        ]
    )
    return tuple(zip(before, after))


def _stack_s_matrices(
    layer_solve_results: Sequence[fmm.LayerSolveResult],
    layer_thicknesses: Sequence[jnp.ndarray],
    force_x64_solve: bool,
) -> Tuple[ScatteringMatrix, ...]:
    """Computes the s-matrices for a stack of layers.

    The matrices are for various substacks: the first scattering matrix is
    for the first layer only, the second is for the stack consisting of the
    first and second layer, etc.

    Args:
        layer_solve_results: The eigensolve results for layers in the stack.
        layer_thicknesses: The thicknesses for layers in the stack.
        force_x64_solve: If `True`, matrix solves will be done with 64 bit precision.

    Returns:
        The tuple of `ScatteringMatrix`.
    """
    if len(layer_solve_results) != len(layer_thicknesses):
        raise ValueError(
            f"`layer_solve_results` and `layer_thicknesses` should have the same "
            f"length but got {len(layer_solve_results)} and {len(layer_thicknesses)}."
        )

    # The initial scattering matrix is just the identity matrix, with the
    # necessary batch dimensions.
    eye = utils.diag(jnp.ones_like(layer_solve_results[0].eigenvalues))
    layer_s_matrix_0 = ScatteringMatrix(
        s11=eye,
        s12=jnp.zeros_like(eye),
        s21=jnp.zeros_like(eye),
        s22=eye,
        start_layer_solve_result=layer_solve_results[0],
        start_layer_thickness=layer_thicknesses[0],
        end_layer_solve_result=layer_solve_results[0],
        end_layer_thickness=layer_thicknesses[0],
    )

    if len(layer_solve_results) == 1:
        return (layer_s_matrix_0,)

    layer_s_matrix_1 = _pair_s_matrix(
        layer_solve_result=layer_solve_results[0],
        layer_thickness=layer_thicknesses[0],
        next_layer_solve_result=layer_solve_results[1],
        next_layer_thickness=layer_thicknesses[1],
        force_x64_solve=force_x64_solve,
    )

    # TODO(mfschubert): Figure out how to use `jax.lax.fori_loop` or similar.
    s_matrices = [layer_s_matrix_0, layer_s_matrix_1]
    for layer_solve_result, layer_thickness in zip(
        layer_solve_results[2:], layer_thicknesses[2:]
    ):
        s_matrices.append(
            append_layer(
                s_matrices[-1],
                layer_solve_result,
                layer_thickness,
                force_x64_solve=force_x64_solve,
            )
        )
    return tuple(s_matrices)


def stack_s_matrix_scan(
    layer_solve_results: fmm.LayerSolveResult,
    layer_thicknesses: jnp.ndarray,
    force_x64_solve: bool = False,
) -> ScatteringMatrix:
    """Computes the scattering matrix for a stack of layers by a scan operation.

    Unlike `stack_s_matrix`, this function uses a scan operation rather than a python
    for loop, which can lead to significantly smaller programs and shorter compile
    times. However, it requires that the layer solve results and thicknesses for be
    represented in the leading batch dimension of the arguments.

    This function is best used when the eigensolve for each layer is also done in a
    batched manner.

    Args:
        layer_solve_results: The layer solve results for all layers in the stack.
        layer_thicknesses: The layer thicknesses for all layers in the stack.
        force_x64_solve: If `True`, matrix solves will be done with 64 bit precision.

    Returns:
        The scattering matrix for the stack.
    """
    assert layer_thicknesses.ndim == 1
    if layer_solve_results.batch_shape[0] != len(layer_thicknesses):
        raise ValueError(
            f"`layer_solve_results` and `layer_thicknesses` should be compatible (i.e. "
            f"correspond to the same number of layers) but inferred layer numbers of "
            f"{layer_solve_results.batch_shape[0]} and {layer_thicknesses.shape[0]}."
        )

    eye = utils.diag(
        jnp.ones(
            layer_solve_results.eigenvalues.shape[1:],
            dtype=layer_solve_results.eigenvalues.dtype,
        )
    )
    start_solve_result = tree_util.tree_map(lambda x: x[0, ...], layer_solve_results)
    s_matrix = ScatteringMatrix(
        s11=eye,
        s12=jnp.zeros_like(eye),
        s21=jnp.zeros_like(eye),
        s22=eye,
        start_layer_solve_result=start_solve_result,
        start_layer_thickness=layer_thicknesses[0],
        end_layer_solve_result=start_solve_result,
        end_layer_thickness=layer_thicknesses[0],
    )

    def scan_fn(s_matrix, x):
        next_layer_solve_result, next_layer_thickness = x
        s_matrix = append_layer(
            s_matrix,
            next_layer_solve_result,
            next_layer_thickness,
            force_x64_solve=force_x64_solve,
        )
        return s_matrix, s_matrix

    s_matrix, _ = jax.lax.scan(
        scan_fn,
        init=s_matrix,
        xs=(
            tree_util.tree_map(lambda x: x[1:], layer_solve_results),
            layer_thicknesses[1:],
        ),
    )
    return s_matrix


def redheffer_star_product(
    a: ScatteringMatrix,
    b: ScatteringMatrix,
    force_x64_solve: bool = False,
) -> ScatteringMatrix:
    """Compute the Redheffer star product of two scattering matrices."""
    a_extended = append_layer(a, b.start_layer_solve_result, b.start_layer_thickness)
    a11, a12, a21, a22 = a_extended.s11, a_extended.s12, a_extended.s21, a_extended.s22
    b11, b12, b21, b22 = b.s11, b.s12, b.s21, b.s22

    _solve = functools.partial(solve, force_x64_solve=force_x64_solve)

    # See https://en.wikipedia.org/wiki/Redheffer_star_product
    eye = utils.diag(jnp.ones_like(a11[..., 0]))
    s11 = b11 @ _solve(eye - a12 @ b21, a11)
    s12 = b12 + b11 @ _solve(eye - a12 @ b21, a12 @ b22)
    s21 = a21 + a22 @ _solve(eye - b21 @ a12, b21 @ a11)
    s22 = a22 @ _solve(eye - b21 @ a12, b22)
    return ScatteringMatrix(
        s11=s11,
        s12=s12,
        s21=s21,
        s22=s22,
        start_layer_solve_result=a.start_layer_solve_result,
        start_layer_thickness=a.start_layer_thickness,
        end_layer_solve_result=b.end_layer_solve_result,
        end_layer_thickness=b.end_layer_thickness,
    )


def append_layer(
    s_matrix: ScatteringMatrix,
    next_layer_solve_result: fmm.LayerSolveResult,
    next_layer_thickness: jnp.ndarray,
    force_x64_solve: bool = False,
) -> ScatteringMatrix:
    """Returns new scattering matrix for the stack with an appended layer.

    Args:
        s_matrix: The existing scattering matrix.
        next_layer_solve_result: The eigensolve result for the layer to append.
        next_layer_thickness: The thickness for the layer to append.
        force_x64_solve: If `True`, matrix solves will be done with 64 bit precision.

    Returns:
        The new `ScatteringMatrix`.
    """
    s11_next, s12_next, s21_next, s22_next = _extend_s_matrix(
        s_matrix_blocks=(s_matrix.s11, s_matrix.s12, s_matrix.s21, s_matrix.s22),
        layer_solve_result=s_matrix.end_layer_solve_result,
        layer_thickness=s_matrix.end_layer_thickness,
        next_layer_solve_result=next_layer_solve_result,
        next_layer_thickness=next_layer_thickness,
        force_x64_solve=force_x64_solve,
    )
    return ScatteringMatrix(
        s11=s11_next,
        s12=s12_next,
        s21=s21_next,
        s22=s22_next,
        start_layer_solve_result=s_matrix.start_layer_solve_result,
        start_layer_thickness=s_matrix.start_layer_thickness,
        end_layer_solve_result=next_layer_solve_result,
        end_layer_thickness=next_layer_thickness,
    )


def prepend_layer(
    s_matrix: ScatteringMatrix,
    next_layer_solve_result: fmm.LayerSolveResult,
    next_layer_thickness: jnp.ndarray,
    force_x64_solve: bool = False,
) -> ScatteringMatrix:
    """Returns new scattering matrix for the stack with a prepended layer.

    Args:
        s_matrix: The existing scattering matrix.
        next_layer_solve_result: The eigensolve result for the layer to append.
        next_layer_thickness: The thickness for the layer to append.
        force_x64_solve: If `True`, matrix solves will be done with 64 bit precision.

    Returns:
        The new `ScatteringMatrix`.
    """
    # To prepend a layer, we compute the scattering matrix that results from
    # appending the layer to the reversed stack. The scattering matrix for
    # the reversed stack is simply the re-ordering of the matrix blocks, i.e.
    # s11 become s22, s12 becomes s21, etc.
    s22_next, s21_next, s12_next, s11_next = _extend_s_matrix(
        s_matrix_blocks=(s_matrix.s22, s_matrix.s21, s_matrix.s12, s_matrix.s11),
        layer_solve_result=s_matrix.start_layer_solve_result,
        layer_thickness=s_matrix.start_layer_thickness,
        next_layer_solve_result=next_layer_solve_result,
        next_layer_thickness=next_layer_thickness,
        force_x64_solve=force_x64_solve,
    )
    return ScatteringMatrix(
        s11=s11_next,
        s12=s12_next,
        s21=s21_next,
        s22=s22_next,
        start_layer_solve_result=next_layer_solve_result,
        start_layer_thickness=next_layer_thickness,
        end_layer_solve_result=s_matrix.end_layer_solve_result,
        end_layer_thickness=s_matrix.end_layer_thickness,
    )


def _extend_s_matrix(
    s_matrix_blocks: Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray],
    layer_solve_result: fmm.LayerSolveResult,
    layer_thickness: jnp.ndarray,
    next_layer_solve_result: fmm.LayerSolveResult,
    next_layer_thickness: jnp.ndarray,
    force_x64_solve: bool,
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Extends the scattering matrix, adding a layer to the end.

    The approach here follows section 5 of [1999 Whittaker].

    Args:
        s_matrix_blocks: The elements `(s11, s12, s21, s22)`.
        layer_solve_result: The eigensolve result of the ending layer.
        layer_thickness: The thickness of the ending layer.
        next_layer_solve_result: The eigensolve result for the layer to append.
        next_layer_thickness: The thickness for the layer to append.
        force_x64_solve: If `True`, matrix solves will be done with 64 bit precision.

    Returns:
        The new `ScatteringMatrix`.
    """
    # Alias for brevity: eigenvalues, eigenvectors, and omega-k matrix.
    q = layer_solve_result.eigenvalues
    phi = layer_solve_result.eigenvectors
    omega_k = layer_solve_result.omega_script_k_matrix

    next_q = next_layer_solve_result.eigenvalues
    next_phi = next_layer_solve_result.eigenvectors
    next_omega_k = next_layer_solve_result.omega_script_k_matrix

    _solve = functools.partial(solve, force_x64_solve=force_x64_solve)

    # Compute the interface matrices following equation 5.3 of [1999 Whittaker].
    # These make use the matrix form of the orthogonality relation of equation 3.9
    # in [1999 Whittaker], i.e. `phi_T @ omega_k @ phi = 1`, to compute `phi_T`.
    # Throughout, we use optimized expressions which avoid matrix inversion and
    # matrix-matrix multiplications. More straightforward implementations are in
    # comments.
    #
    # phi_T = jnp.linalg.inv(omega_k @ phi)
    # term1 = diag(q) @ phi_T @ next_omega_k @ next_phi @ diag(1 / next_q)
    term1 = q[..., jnp.newaxis] * _solve(
        omega_k @ phi,
        next_omega_k @ next_phi * (1 / next_q)[..., jnp.newaxis, :],
    )
    # term2 = phi_T @ omega_k @ next_phi
    term2 = _solve(omega_k @ phi, omega_k @ next_phi)
    i11 = i22 = 0.5 * (term1 + term2)
    i12 = i21 = 0.5 * (-term1 + term2)

    # Phase terms \hat{f}(d) defined near equation 4.2 of [1999 Whittaker]. These
    # describe phase accumulated by propagating across a layer for each eigenmode.
    fd = jnp.exp(1j * q * layer_thickness)
    fd_next = jnp.exp(1j * next_q * next_layer_thickness)

    # Update the s-matrix to include the present layer, following the recipe
    # given in equation 5.4 of [1999 Whittaker].
    s11, s12, s21, s22 = s_matrix_blocks

    # s11_next = inv(i11 - diag(fd) @ s12 @ i21) @ diag(fd) @ s11
    term3 = i11 - fd[..., jnp.newaxis] * s12 @ i21
    s11_next = _solve(term3, fd[..., jnp.newaxis] * s11)
    # s12_next = inv(i11 - diag(fd) @ s12 @ i21)
    #            @ (diag(fd) @ s12 @ i22 - i12) @ diag(fd_next)
    s12_next = _solve(
        term3,
        (fd[..., jnp.newaxis] * s12 @ i22 - i12) * fd_next[..., jnp.newaxis, :],
    )
    s21_next = s22 @ i21 @ s11_next + s21
    # s22_next = s22 @ i21 @ s12_next + s22 @ i22 @ diag(fd_next)
    s22_next = s22 @ i21 @ s12_next + s22 @ i22 * fd_next[..., jnp.newaxis, :]

    return (s11_next, s12_next, s21_next, s22_next)


def _pair_s_matrix(
    layer_solve_result: fmm.LayerSolveResult,
    layer_thickness: jnp.ndarray,
    next_layer_solve_result: fmm.LayerSolveResult,
    next_layer_thickness: jnp.ndarray,
    force_x64_solve: bool,
) -> ScatteringMatrix:
    """Generate the scattering matrix for a pair of layers."""
    # Alias for brevity: eigenvalues, eigenvectors, and omega-k matrix.
    q = layer_solve_result.eigenvalues
    phi = layer_solve_result.eigenvectors
    omega_k = layer_solve_result.omega_script_k_matrix

    next_q = next_layer_solve_result.eigenvalues
    next_phi = next_layer_solve_result.eigenvectors
    next_omega_k = next_layer_solve_result.omega_script_k_matrix

    _solve = functools.partial(solve, force_x64_solve=force_x64_solve)

    term1 = q[..., jnp.newaxis] * _solve(
        omega_k @ phi,
        next_omega_k @ next_phi * (1 / next_q)[..., jnp.newaxis, :],
    )
    term2 = _solve(omega_k @ phi, omega_k @ next_phi)
    i11 = i22 = 0.5 * (term1 + term2)
    i12 = i21 = 0.5 * (-term1 + term2)

    fd = jnp.exp(1j * q * layer_thickness)
    fd_next = jnp.exp(1j * next_q * next_layer_thickness)

    # The computation is identical to that in `_extend_s_matrix` with `s11` and `s22`
    # being the identity, and `s12` and `s21` being zero.
    fd_diag = jnp.expand_dims(utils.diag(fd), tuple(range(i11.ndim - fd.ndim - 1)))
    s11 = _solve(i11, fd_diag)
    s12 = _solve(i11, -i12 * fd_next[..., jnp.newaxis, :])
    s21 = i21 @ s11
    s22 = i21 @ s12 + i22 * fd_next[..., jnp.newaxis, :]

    return ScatteringMatrix(
        s11=s11,
        s12=s12,
        s21=s21,
        s22=s22,
        start_layer_solve_result=layer_solve_result,
        start_layer_thickness=layer_thickness,
        end_layer_solve_result=next_layer_solve_result,
        end_layer_thickness=next_layer_thickness,
    )


def set_end_layer_thickness(
    s_matrix: ScatteringMatrix,
    thickness: jnp.ndarray,
) -> ScatteringMatrix:
    """Returns a new `ScatteringMatrix` with a modified end layer thickness.

    Args:
        s_matrix: The initial `ScatteringMatrix`.
        thickness: The desired thickness of the layer.

    Returns:
        The new `ScatteringMatrix`.
    """
    q = s_matrix.end_layer_solve_result.eigenvalues
    fd = jnp.exp(1j * q * (thickness - s_matrix.end_layer_thickness))
    return ScatteringMatrix(
        s11=s_matrix.s11,
        s12=s_matrix.s12 * fd[..., jnp.newaxis, :],
        s21=s_matrix.s21,
        s22=s_matrix.s22 * fd[..., jnp.newaxis, :],
        start_layer_solve_result=s_matrix.start_layer_solve_result,
        start_layer_thickness=s_matrix.start_layer_thickness,
        end_layer_solve_result=s_matrix.end_layer_solve_result,
        end_layer_thickness=thickness,
    )


def set_start_layer_thickness(
    s_matrix: ScatteringMatrix,
    thickness: jnp.ndarray,
) -> ScatteringMatrix:
    """Returns a new `ScatteringMatrix` with a modified start layer thickness.

    Args:
        s_matrix: The initial `ScatteringMatrix`.
        thickness: The desired thickness of the layer.

    Returns:
        The new `ScatteringMatrix`.
    """
    q = s_matrix.start_layer_solve_result.eigenvalues
    fd = jnp.exp(1j * q * (thickness - s_matrix.start_layer_thickness))
    return ScatteringMatrix(
        s11=s_matrix.s11 * fd[..., jnp.newaxis, :],
        s12=s_matrix.s12,
        s21=s_matrix.s21 * fd[..., jnp.newaxis, :],
        s22=s_matrix.s22,
        start_layer_solve_result=s_matrix.start_layer_solve_result,
        start_layer_thickness=thickness,
        end_layer_solve_result=s_matrix.end_layer_solve_result,
        end_layer_thickness=s_matrix.end_layer_thickness,
    )


# -----------------------------------------------------------------------------
# Register custom objects in this module with jax to enable `jit`.
# -----------------------------------------------------------------------------


tree_util.register_pytree_node(
    ScatteringMatrix,
    lambda x: (
        (
            x.s11,
            x.s12,
            x.s21,
            x.s22,
            x.start_layer_solve_result,
            x.start_layer_thickness,
            x.end_layer_solve_result,
            x.end_layer_thickness,
        ),
        None,
    ),
    lambda _, x: ScatteringMatrix(*x),
)


def solve(a: jnp.ndarray, b: jnp.ndarray, *, force_x64_solve: bool) -> jnp.ndarray:
    """Solves `A @ x = b`, optionally using 64-bit precision."""
    output_dtype = jnp.promote_types(a.dtype, b.dtype)
    if force_x64_solve and jax.config.read("jax_enable_x64"):
        a = a.astype(jnp.promote_types(a.dtype, jnp.float64))
        b = b.astype(jnp.promote_types(b.dtype, jnp.float64))
    return jnp.linalg.solve(a, b).astype(output_dtype)



# Copyright (c) Meta Platforms, Inc. and affiliates.

__version__ = "v0.12.0"

from fmmax import basis as basis
from fmmax import beams as beams
from fmmax import farfield as farfield
from fmmax import fft as fft
from fmmax import fields as fields
from fmmax import fmm as fmm
from fmmax import fmm_matrices as fmm_matrices
from fmmax import scattering as scattering
from fmmax import sources as sources
from fmmax import translate as translate
from fmmax import utils as utils
from fmmax import vector as vector



"""Functions related to perfectly matched layers.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import dataclasses
from typing import Tuple

import jax.numpy as jnp
from jax import tree_util


@dataclasses.dataclass
class PMLParams:
    """Stores parameters that define perfectly matched layers.

    Attributes:
        num_x: The number of grid points occupied by the PML, in the x-direction.
        num_y: The number of grid points occupied by the PML, in the y-direction.
        a_max: The strength parameter for uniaxial PML.
        p: The exponent parameter for uniaxial PML.
        sigma_max: The conductivity parameter for uniaxial PML.
    """

    num_x: int
    num_y: int
    a_max: float = 4.0
    p: float = 4.0
    sigma_max: float = 1.0


def apply_uniaxial_pml(
    permittivity: jnp.ndarray,
    pml_params: PMLParams,
) -> Tuple[
    Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray],
    Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray],
]:
    """Generate the permittivity and permeability tensor elements for uniaxial pml.

    The PML assumes that the unit cell has primitive lattice vectors u and v
    which are parallel to x and y axes, respectively.

    This function is appropriate for isotropic nonmagnetic media, but the
    permittivities and permeabilities generated are anisotropic.

    Args:
        permittivity: isotropic permittivity
        pml_params: The parameters defining the perfectly matched layer dimensions
            and absorption characteristics.

    Returns:
        The permittivity and permeability tensor elements,
        `(
            (eps_xx, eps_xy, eps_yx, eps_yy, eps_zz),
            (mu_xx, mu_xy, mu_yx, mu_yy, mu_zz),
        )`.
    """
    permittivity = _crop_and_edge_pad_pml_region(
        permittivity, widths=(pml_params.num_x, pml_params.num_y)
    )

    dx, dy = _normalized_distance_into_pml(
        shape=permittivity.shape[-2:],  # type: ignore[arg-type]
        widths=(pml_params.num_x, pml_params.num_y),
    )

    sx = (1 + pml_params.a_max * dx**pml_params.p) * (
        1 + 1j * pml_params.sigma_max * jnp.sin(jnp.pi / 2 * dx) ** 2
    )
    sy = (1 + pml_params.a_max * dy**pml_params.p) * (
        1 + 1j * pml_params.sigma_max * jnp.sin(jnp.pi / 2 * dy) ** 2
    )
    sz = 1

    permittivity_xx = sy * sz / sx * permittivity
    permittivity_yy = sx * sz / sy * permittivity
    permittivity_zz = sx * sy / sz * permittivity
    permittivity_xy = jnp.zeros_like(permittivity)
    permittivity_yx = jnp.zeros_like(permittivity)

    permeability_xx = sy * sz / sx * jnp.ones_like(permittivity)
    permeability_yy = sx * sz / sy * jnp.ones_like(permittivity)
    permeability_zz = sx * sy / sz * jnp.ones_like(permittivity)
    permeability_xy = jnp.zeros_like(permittivity)
    permeability_yx = jnp.zeros_like(permittivity)

    return (
        permittivity_xx,
        permittivity_xy,
        permittivity_yx,
        permittivity_yy,
        permittivity_zz,
    ), (
        permeability_xx,
        permeability_xy,
        permeability_yx,
        permeability_yy,
        permeability_zz,
    )


def _crop_and_edge_pad_pml_region(
    permittivity: jnp.ndarray,
    widths: Tuple[int, int],
) -> jnp.ndarray:
    """Crops the trailing dimensions of `permittivity` and applies edge padding."""
    i_width, j_width = widths
    if (i_width * 2, j_width * 2) >= permittivity.shape[-2:]:
        raise ValueError(
            f"`widths` {widths} are incompatible with permittivity shape "
            f"{permittivity.shape}."
        )

    arr_cropped = permittivity[
        ...,
        i_width : permittivity.shape[-2] - i_width,
        j_width : permittivity.shape[-1] - j_width,
    ]

    pad_width = ((0, 0),) * (permittivity.ndim - 2) + (
        (i_width, i_width),
        (j_width, j_width),
    )
    return jnp.pad(arr_cropped, pad_width, mode="edge")


def _normalized_distance_into_pml(
    shape: Tuple[int, int],
    widths: Tuple[int, int],
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Compute the distance into the PML layer, in terms of array elements."""
    i, j = jnp.meshgrid(
        jnp.arange(shape[-2]),
        jnp.arange(shape[-1]),
        indexing="ij",
    )
    i_width, j_width = widths

    i_distance = jnp.maximum(i_width - i, 0)
    i_distance = jnp.maximum(i_distance, i - (shape[-2] - i_width - 1))

    j_distance = jnp.maximum(j_width - j, 0)
    j_distance = jnp.maximum(j_distance, j - (shape[-1] - j_width - 1))

    i_scale = jnp.where(widths[0] == 0, 1, 1 / jnp.where(widths[0] == 0, 1, widths[0]))
    j_scale = jnp.where(widths[1] == 0, 1, 1 / jnp.where(widths[1] == 0, 1, widths[1]))
    return i_distance * i_scale, j_distance * j_scale


# -----------------------------------------------------------------------------
# Register custom objects in this module with jax to enable `jit`.
# -----------------------------------------------------------------------------


tree_util.register_pytree_node(
    PMLParams,
    lambda p: ((), (p.num_x, p.num_y, p.a_max, p.p, p.sigma_max)),
    lambda values, _: PMLParams(*values),
)



"""Functions related to eigensolve with translated permittivity distributions.

Copyright (c) 2024 Martin F. Schubert
"""

import functools

import jax.numpy as jnp

from fmmax import basis, fft, fmm


def translate_layer_solve_result(
    solve_result: fmm.LayerSolveResult,
    dx: jnp.ndarray,
    dy: jnp.ndarray,
) -> fmm.LayerSolveResult:
    """Obtain solve result for layer with translated permittivity distribution.

    Rather than performing eigensolve on translated permittivity distribution, this
    function allows one to translate the solve result. This can result in significant
    compute savings for structures such as slanted gratings.

    Args:
        solve_result: The `LayerSolveResult` for the permittivity distribution prior
            to translation.
        dx: The translation of the permittivity distribution along the `x` direction.
        dy: The translation of the permittivity distribution along the `y` direction.

    Returns:
        The `LayerSolveResult` for the layer with translated permittivity.
    """
    assert dx.shape == dy.shape == ()

    _shift = functools.partial(
        _apply_shift_eigenvectors,
        expansion=solve_result.expansion,
        primitive_lattice_vectors=solve_result.primitive_lattice_vectors,
        dx=dx,
        dy=dy,
    )

    ev_te, ev_tm = jnp.split(solve_result.eigenvectors, 2, axis=-2)
    shifted_eigenvectors = jnp.concatenate([_shift(ev_te), _shift(ev_tm)], axis=-2)

    _shift_toeplitz = functools.partial(
        _apply_shift_toeplitz,
        expansion=solve_result.expansion,
        primitive_lattice_vectors=solve_result.primitive_lattice_vectors,
        dx=dx,
        dy=dy,
    )

    m0, m1 = jnp.split(solve_result.transverse_permeability_matrix, 2, axis=-2)
    m00, m01 = jnp.split(m0, 2, axis=-1)
    m10, m11 = jnp.split(m1, 2, axis=-1)
    shifted_transverse_permeability_matrix = jnp.block(
        [
            [_shift_toeplitz(m00), _shift_toeplitz(m01)],
            [_shift_toeplitz(m10), _shift_toeplitz(m11)],
        ],
    )

    return fmm.LayerSolveResult(
        wavelength=solve_result.wavelength,
        in_plane_wavevector=solve_result.in_plane_wavevector,
        primitive_lattice_vectors=solve_result.primitive_lattice_vectors,
        expansion=solve_result.expansion,
        eigenvalues=solve_result.eigenvalues,
        eigenvectors=shifted_eigenvectors,
        z_permittivity_matrix=_shift_toeplitz(solve_result.z_permittivity_matrix),
        inverse_z_permittivity_matrix=_shift_toeplitz(
            solve_result.inverse_z_permittivity_matrix
        ),
        z_permeability_matrix=_shift_toeplitz(solve_result.z_permeability_matrix),
        inverse_z_permeability_matrix=_shift_toeplitz(
            solve_result.inverse_z_permeability_matrix
        ),
        transverse_permeability_matrix=shifted_transverse_permeability_matrix,
        tangent_vector_field=solve_result.tangent_vector_field,
    )


def _apply_shift_eigenvectors(
    eigenvectors: jnp.ndarray,
    expansion: basis.Expansion,
    primitive_lattice_vectors: basis.LatticeVectors,
    dx: jnp.ndarray,
    dy: jnp.ndarray,
) -> jnp.ndarray:
    """Returns eigenvectors that result from translating a structure."""
    transverse_wavevectors = basis.transverse_wavevectors(
        in_plane_wavevector=jnp.zeros((2,)),
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
    )
    kx = transverse_wavevectors[:, 0]
    ky = transverse_wavevectors[:, 1]
    return eigenvectors * jnp.exp(-1j * (kx * dx + ky * dy))[..., jnp.newaxis]


def _apply_shift_toeplitz(
    mat: jnp.ndarray,
    expansion: basis.Expansion,
    primitive_lattice_vectors: basis.LatticeVectors,
    dx: jnp.ndarray,
    dy: jnp.ndarray,
) -> jnp.ndarray:
    """Return toeplitz matrix that results from translating a structure."""
    idx = fft._standard_toeplitz_indices(expansion)

    reciprocal_vectors = primitive_lattice_vectors.reciprocal
    u = reciprocal_vectors.u
    v = reciprocal_vectors.v
    kx = 2 * jnp.pi * (idx[:, :, 0] * u[0] + idx[:, :, 1] * v[0])
    ky = 2 * jnp.pi * (idx[:, :, 0] * u[1] + idx[:, :, 1] * v[1])
    return mat * jnp.exp(-1j * (kx * dx + ky * dy))



"""Defines functions related to farfield patterns in the FMM scheme.

These functions are intended for calculations that involve Brillouin
zone integration, and batched in-plane wavevectors as generated by
`basis.brillouin_zone_in_plane_wavevector`.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

from typing import Callable, Tuple

import jax
import jax.numpy as jnp
import numpy as onp

from fmmax import basis, utils


def farfield_profile(
    flux: jnp.ndarray,
    wavelength: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    expansion: basis.Expansion,
    brillouin_grid_axes: Tuple[int, int],
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Computes a farfield profile.

    This function effectively "unstacks" the values for each Fourier order and
    for each point in the Brillouin zone sampling scheme.

    Args:
        flux: The flux array, with shape `(..., num_bz_kx, num_bz_ky, ...
            2 * num_terms, num_sources)`.
        wavelength: The wavelength, batch-compatible with `flux`.
        in_plane_wavevector: The in-plane wavevector for the zeroth Fourier
            order, batch-compatible with `flux`.
        primitive_lattice_vectors: The primitive lattice vectors of the unit cell.
        expansion: The expansion used for the fields.
        brillouin_grid_axes: Specifies the two axes of `flux` corresponding to
            the Brillouin zone grid.

    Returns:
        The polar and azimuthal angles, solid angle associated with each value,
        and the farfield power.
    """
    assert flux.shape[-2] == 2 * expansion.num_terms
    brillouin_grid_axes = utils.absolute_axes(  # type: ignore[assignment]
        brillouin_grid_axes, flux.ndim
    )

    ndim_batch = flux.ndim - 2
    wavelength = utils.atleast_nd(wavelength, ndim_batch)
    in_plane_wavevector = utils.atleast_nd(in_plane_wavevector, ndim_batch + 1)

    transverse_wavevectors = basis.transverse_wavevectors(
        in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
    )
    transverse_wavevectors = unflatten_transverse_wavevectors(
        transverse_wavevectors, expansion, brillouin_grid_axes
    )
    flux = unflatten_flux(flux, expansion, brillouin_grid_axes)

    # Remove the brillouin zone axes from `wavelength`, making it compatible
    # with the unflattened flux and transverse wavevectors.
    wavelength = jnp.squeeze(wavelength, axis=brillouin_grid_axes)

    polar_angle, azimuthal_angle = angles_from_unflattened_transverse_wavevectors(
        transverse_wavevectors=transverse_wavevectors,
        wavelength=wavelength,
    )

    # Transform flux form units of power per unit Brillouin zone area to
    # power per unit solid angle. Add dummy dimensions for the polarization
    # and sources.
    solid_angle = solid_angle_from_unflattened_transverse_wavevectors(
        transverse_wavevectors=transverse_wavevectors,
        wavelength=wavelength,
    )
    transformed_flux = flux / solid_angle[..., jnp.newaxis, jnp.newaxis]
    return polar_angle, azimuthal_angle, solid_angle, transformed_flux


def angles_from_unflattened_transverse_wavevectors(
    transverse_wavevectors: jnp.ndarray,
    wavelength: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Computes the propagation angles in free space for given wavevectors.

    Evanescent modes whose transverse wavevector magnitude exceeds that of
    the free space wavevector are given a polar angle of `pi / 2`.

    Args:
        transverse_wavevectors: The unflattened transverse wavectors, with
            shape `(..., nkx, nky, 2)`.
        wavelength: The free-space wavelength.

    Returns:
        Arrays containing the polar and azimuthal angles.
    """
    assert transverse_wavevectors.ndim - 3 == wavelength.ndim
    kx = transverse_wavevectors[..., 0]
    ky = transverse_wavevectors[..., 1]
    kt = jnp.sqrt(kx**2 + ky**2)

    sin_polar_angle = kt * wavelength[..., jnp.newaxis, jnp.newaxis] / (2 * jnp.pi)
    polar_angle = jnp.where(
        sin_polar_angle > 1, jnp.pi / 2, jnp.arcsin(sin_polar_angle)
    )

    azimuthal_angle = jnp.angle(kx + 1j * ky)
    return polar_angle, azimuthal_angle


def solid_angle_from_unflattened_transverse_wavevectors(
    transverse_wavevectors: jnp.ndarray,
    wavelength: jnp.ndarray,
) -> jnp.ndarray:
    """Computes the solid angle associated with each transverse wavevector.

    The transverse wavevectors should be unflattened, i.e. the `-3` and `-2`
    axes should correspond to different points in k-space.

    Args:
        transverse_wavevectors: The unflattened transverse wavevectors
        wavelength: The free-space wavelength.

    Returns:
        The solid angle, with the shape matching the leading dimensions of
        `transverse_wavevectors`.
    """
    assert transverse_wavevectors.ndim >= 3

    kx = transverse_wavevectors[..., 0]
    ky = transverse_wavevectors[..., 1]

    # Normalize the wavevectors so they lie on the unit sphere.
    kx /= 2 * jnp.pi / wavelength[..., jnp.newaxis, jnp.newaxis]
    ky /= 2 * jnp.pi / wavelength[..., jnp.newaxis, jnp.newaxis]
    polar_angle = jnp.arcsin(jnp.sqrt(kx**2 + ky**2))

    # Each of our transverse wavevectors lies within a "cell" in the kxky plane.
    # Compute the area of each cell, and then project it onto the unit sphere
    # to get the solid angle associated with each transverse wavevector.
    #
    # First, compute the locations of the cell verteces.
    pad_width = ((0, 0),) * (kx.ndim - 2) + ((1, 1), (1, 1))
    vertex_kx = jnp.pad(kx, pad_width, mode="edge")
    vertex_ky = jnp.pad(ky, pad_width, mode="edge")
    vertex_kx = (
        vertex_kx[..., :-1, :-1]
        + vertex_kx[..., :-1, 1:]
        + vertex_kx[..., 1:, :-1]
        + vertex_kx[..., 1:, 1:]
    ) / 4
    vertex_ky = (
        vertex_ky[..., :-1, :-1]
        + vertex_ky[..., :-1, 1:]
        + vertex_ky[..., 1:, :-1]
        + vertex_ky[..., 1:, 1:]
    ) / 4
    vertex_kt = jnp.stack([vertex_kx, vertex_ky], axis=-1)

    # Find the vectors defining each parallelogramic cell.
    v1 = vertex_kt[..., :-1, 1:, :] - vertex_kt[..., :-1, :-1, :]
    v2 = vertex_kt[..., 1:, :-1, :] - vertex_kt[..., :-1, :-1, :]

    # Find the area of each parallelogramic cell.
    cell_area = jnp.abs(v1[..., 0] * v2[..., 1] - v2[..., 0] * v1[..., 1])

    # Project the area onto the unit sphere.
    projected_area = cell_area / jnp.cos(polar_angle)
    return projected_area


# -----------------------------------------------------------------------------
# Functions for computing the total flux in some angular cone.
# -----------------------------------------------------------------------------


def integrated_flux(
    flux: jnp.ndarray,
    wavelength: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    expansion: basis.Expansion,
    brillouin_grid_axes: Tuple[int, int],
    angle_bounds_fn: Callable[[jnp.ndarray, jnp.ndarray], jnp.ndarray],
    upsample_factor: int,
) -> jnp.ndarray:
    """Computes the flux within the bounds defined by `angle_bounds_fn`.

    Args:
        flux: The flux array, with shape `(..., num_bz_kx, num_bz_ky, ...
            2 * num_terms, num_sources)`.
        wavelength: The wavelength, batch-compatible with `flux`.
        in_plane_wavevector: The in-plane wavevector for the zeroth Fourier
            order, batch-compatible with `flux`.
        primitive_lattice_vectors: The primitive lattice vectors of the unit cell.
        expansion: The expansion used for the fields.
        brillouin_grid_axes: Specifies the two axes of `flux` corresponding to
            the Brillouin zone grid.
        angle_bounds_fn: A function with signature `fn(polar_angle, azimuthal_angle)`
            returning a mask that is `True` for angles that should be included in
            the integral.
        upsample_factor: Integer factor specifying upsampling performed in the
            integral, which is used to approximate trapezoidal rule integration.

    Returns:
        The integrated flux, with shape equal to the batch dimensions of flux,
        excluding those for the brillouin zone grid.
    """
    brillouin_grid_axes = utils.absolute_axes(  # type: ignore[assignment]
        brillouin_grid_axes, flux.ndim
    )

    # Compute the weights array, which reduce the integration weights to
    # an inner product.
    weights = _integrated_flux_weights(
        flux=flux,
        wavelength=wavelength,
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
        brillouin_grid_axes=brillouin_grid_axes,
        angle_bounds_fn=angle_bounds_fn,
        upsample_factor=upsample_factor,
    )

    # Sum over the Brillouin zone and Fourier order axes.
    return jnp.sum(weights * flux, axis=brillouin_grid_axes + (-2,))


def _integrated_flux_weights(
    flux: jnp.ndarray,
    wavelength: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    expansion: basis.Expansion,
    brillouin_grid_axes: Tuple[int, int],
    angle_bounds_fn: Callable[[jnp.ndarray, jnp.ndarray], jnp.ndarray],
    upsample_factor: int,
) -> jnp.ndarray:
    """Returns the integration weights for the bounds defined by `angle_bounds_fn`."""
    brillouin_grid_axes = utils.absolute_axes(  # type: ignore[assignment]
        brillouin_grid_axes, flux.ndim
    )

    def _integrated_fn(flux):
        assert flux.shape[-1] == 1
        return jnp.sum(
            _integrated_flux_upsampled(
                flux=flux,
                wavelength=wavelength,
                in_plane_wavevector=in_plane_wavevector,
                primitive_lattice_vectors=primitive_lattice_vectors,
                expansion=expansion,
                brillouin_grid_axes=brillouin_grid_axes,
                angle_bounds_fn=angle_bounds_fn,
                upsample_factor=upsample_factor,
            )
        )

    # The weights are just the gradient of the integrated flux with respect
    # to the flat array elements; use a dummy flux having a single source.
    dummy_flux = jnp.ones_like(flux, shape=flux.shape[:-1] + (1,))
    return jax.grad(_integrated_fn)(dummy_flux)


def _integrated_flux_upsampled(
    flux: jnp.ndarray,
    wavelength: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    expansion: basis.Expansion,
    brillouin_grid_axes: Tuple[int, int],
    angle_bounds_fn: Callable[[jnp.ndarray, jnp.ndarray], jnp.ndarray],
    upsample_factor: int,
) -> jnp.ndarray:
    """Computes the flux within the bounds defined by `angle_bounds_fn`."""
    assert upsample_factor >= 1
    brillouin_grid_axes = utils.absolute_axes(  # type: ignore[assignment]
        brillouin_grid_axes, flux.ndim
    )

    ndim_batch = flux.ndim - 2
    wavelength = utils.atleast_nd(wavelength, ndim_batch)
    in_plane_wavevector = utils.atleast_nd(in_plane_wavevector, ndim_batch + 1)

    transverse_wavevectors = basis.transverse_wavevectors(
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
    )

    flux = unflatten_flux(flux, expansion, brillouin_grid_axes)
    transverse_wavevectors = unflatten_transverse_wavevectors(
        transverse_wavevectors, expansion, brillouin_grid_axes
    )

    # Remove the `nan`s that are found at array locations having no associated
    # value in the original flattened arrays.
    flux = jnp.where(jnp.isnan(flux), 0, flux)
    transverse_wavevectors = jnp.where(
        jnp.isnan(transverse_wavevectors), 0, transverse_wavevectors
    )

    # Upsample the transverse wavevectors and flux to high resolution.
    assert flux.shape[-2] == 2
    upsampled_flux_shape = flux.shape[:-4] + (
        upsample_factor * flux.shape[-4],  # kx axis
        upsample_factor * flux.shape[-3],  # ky axis
        flux.shape[-2],  # Polarization axis, length 2
        flux.shape[-1],  # Source axis
    )
    flux = jax.image.resize(flux, upsampled_flux_shape, method="linear")

    assert transverse_wavevectors.shape[-1] == 2
    upsampled_wavevector_shape = transverse_wavevectors.shape[:-3] + (
        upsample_factor * transverse_wavevectors.shape[-3],  # kx axis
        upsample_factor * transverse_wavevectors.shape[-2],  # ky axis
        transverse_wavevectors.shape[-1],  # Direction axis.
    )
    transverse_wavevectors = jax.image.resize(
        transverse_wavevectors, upsampled_wavevector_shape, method="linear"
    )

    # Remove the brillouin grid axes from wavelength, and insert axes
    # for the unstacked wavevector.
    wavelength = jnp.squeeze(wavelength, brillouin_grid_axes)

    # Compute polar and azimuthal angles.
    polar_angle, azimuthal_angle = angles_from_unflattened_transverse_wavevectors(
        transverse_wavevectors, wavelength
    )

    selected = angle_bounds_fn(polar_angle, azimuthal_angle)
    selected = jnp.where(jnp.isnan(selected), False, selected)
    selected = selected[..., jnp.newaxis, jnp.newaxis]

    masked_flux = jnp.where(selected, flux, 0)

    # Sum over the kx, ky, and polarization axes.
    return jnp.sum(masked_flux, axis=(-4, -3, -2)) / upsample_factor**2


# -----------------------------------------------------------------------------
# Functions related to unflattening results of Brillouin zone integration.
# -----------------------------------------------------------------------------


def unflatten(flat: jnp.ndarray, expansion: basis.Expansion) -> jnp.ndarray:
    """Unflattens an array for a given expansion and Brillouin integration scheme.

    The returned array combines the values associated with all terms in the
    Fourier expansion at all points in the Brillouin zone grid in a single
    array with trailing axes havving shape `(num_kx, num_ky)`. Elements in the
    output which have no corresponding elements in `flat` are given a value
    of `nan`.

    The flat array should have shape `(..., num_bz_kx, num_bz_ky, num_terms)`,
    where `num_terms` is the number of terms in the Fourier expansion, and the
    `-3` and `-2` axes are for the Brillouin zone grid, as used e.g. with
    Brillouin zone integration to model localized sources.

    This function assumes that the Brillouin zone is sampled on a regular grid,
    as produced by `basis.brillouin_zone_in_plane_wavevector`.

    Args:
        flat: The flat array, with shape  `(..., num_bz_kx, num_bz_ky, num_terms)`.
        expansion: The expansion used for the array.

    Returns:
        The unflattened array, with shape `(batch_shape, num_kx, num_ky)`.
    """
    assert flat.ndim >= 3
    assert flat.shape[-1] == expansion.num_terms

    i = expansion.basis_coefficients[:, 0]
    j = expansion.basis_coefficients[:, 1]

    batch_shape = flat.shape[:-3]
    bz_grid_shape = flat.shape[-3:-1]

    # The shape of the output array shoudl accomodate all `(i, j)` values.
    shape = batch_shape + (
        (max(i) - min(i) + 1) * bz_grid_shape[0],
        (max(j) - min(j) + 1) * bz_grid_shape[1],
    )

    bz_i, bz_j = onp.meshgrid(
        onp.arange(bz_grid_shape[0]),
        onp.arange(bz_grid_shape[1]),
        indexing="ij",
    )

    offset_i = min(i) * bz_grid_shape[0]
    offset_j = min(j) * bz_grid_shape[1]

    merged_i = i * bz_grid_shape[0] + bz_i[..., onp.newaxis] - offset_i
    merged_j = j * bz_grid_shape[1] + bz_j[..., onp.newaxis] - offset_j

    stacked_i = merged_i.flatten()
    stacked_j = merged_j.flatten()
    stacked_flat = jnp.reshape(flat, batch_shape + (-1,))

    return jnp.full(shape, jnp.nan).at[..., stacked_i, stacked_j].set(stacked_flat)


def unflatten_flux(
    flux: jnp.ndarray,
    expansion: basis.Expansion,
    brillouin_grid_axes: Tuple[int, int],
) -> jnp.ndarray:
    """Unflattens a flux for a given expansion and Brillouin integration scheme.

    Args:
        flux: The flux array, with shape `(..., num_bz_kx, num_bz_ky, ...
            2 * num_terms, num_sources)`.
        expansion: The expansion used for the flux.
        brillouin_grid_axes: The axes associated with the Brillouin zone grid.

    Returns:
        The unflattened flux, with shape `(..., num_kx, num_ky, 2, num_sources)`.
    """
    assert flux.ndim >= 4
    assert flux.shape[-2] == 2 * expansion.num_terms
    brillouin_grid_axes = utils.absolute_axes(  # type: ignore[assignment]
        brillouin_grid_axes, flux.ndim
    )

    # The flux array has values for two polarizations at each Fourier order. Split
    # these and treat them as a batch dimension.
    flux = jnp.reshape(flux, flux.shape[:-2] + (2, -1, flux.shape[-1]))

    # Transpose so the axes associated with the Fourier orders and Brillouin zone
    # grid are the trailing axes, as needed by `unflatten`.
    batch_axes = tuple(
        [i for i in range(flux.ndim) if i not in brillouin_grid_axes + (flux.ndim - 2,)]
    )
    axes = batch_axes + tuple([i for i in range(flux.ndim) if i not in batch_axes])
    flux = jnp.transpose(flux, axes)

    flux = unflatten(flux, expansion)

    # Transpose so that polarization and sources are returned to the trailing axes.
    axes = tuple(range(flux.ndim - 4)) + (-2, -1, -4, -3)
    return jnp.transpose(flux, axes)


def unflatten_transverse_wavevectors(
    transverse_wavevectors: jnp.ndarray,
    expansion: basis.Expansion,
    brillouin_grid_axes: Tuple[int, int],
) -> jnp.ndarray:
    """Unflatten transverse wavevectors for given expansion and integration scheme.

    Args:
        transverse_wavevectors: The transverse wavevectors array, with shape
            `(..., num_bz_kx, num_bz_ky, ..., num_terms, 2)`.
        expansion: The expansion used for the flux.
        brillouin_grid_axes: The axes associated with the Brillouin zone grid.

    Returns:
        The unflattened wavevectors, with shape `(..., num_kx, num_ky, 2)`.
    """
    assert transverse_wavevectors.ndim >= 4
    assert transverse_wavevectors.shape[-2:] == (expansion.num_terms, 2)
    brillouin_grid_axes = utils.absolute_axes(  # type: ignore[assignment]
        brillouin_grid_axes, transverse_wavevectors.ndim
    )

    # Transpose so the axes associated with the Fourier orders and Brillouin zone
    # grid are the trailing axes, as needed by `unflatten`.
    non_batch_axes = brillouin_grid_axes + (transverse_wavevectors.ndim - 2,)
    batch_axes = tuple(
        [i for i in range(transverse_wavevectors.ndim) if i not in non_batch_axes]
    )
    axes = batch_axes + tuple(
        [i for i in range(transverse_wavevectors.ndim) if i not in batch_axes]
    )
    transverse_wavevectors = jnp.transpose(transverse_wavevectors, axes)

    transverse_wavevectors = unflatten(transverse_wavevectors, expansion)

    # Transpose so the trailing axis is for the wavevector direction.
    axes = tuple(range(transverse_wavevectors.ndim - 3)) + (-2, -1, -3)
    return jnp.transpose(transverse_wavevectors, axes)



"""Functions that generate various matrices for the FMM problem.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import functools
from typing import Tuple

import jax.numpy as jnp

from fmmax import basis, fft, utils


def omega_script_k_matrix_patterned(
    wavelength: jnp.ndarray,
    z_permittivity_matrix: jnp.ndarray,
    transverse_permeability_matrix: jnp.ndarray,
    transverse_wavevectors: jnp.ndarray,
) -> jnp.ndarray:
    """Compute the omega-script-k matrix of equation 26 from [2012 Liu]."""
    # The script-k matrix from equation 19 of [2012 Liu].
    script_k_matrix = script_k_matrix_patterned(
        z_permittivity_matrix, transverse_wavevectors
    )
    angular_frequency = utils.angular_frequency_for_wavelength(wavelength)
    angular_frequency_squared = angular_frequency[..., jnp.newaxis, jnp.newaxis] ** 2
    return angular_frequency_squared * transverse_permeability_matrix - script_k_matrix


# -----------------------------------------------------------------------------
# Functions to compute the k- and script-k matrices.
# -----------------------------------------------------------------------------


def script_k_matrix_uniform(
    permittivity: jnp.ndarray,
    transverse_wavevectors: jnp.ndarray,
) -> jnp.ndarray:
    """Returns the uniform-layer script-k matrix from eq. 19 of [2012 Liu]."""
    kx = transverse_wavevectors[..., 0]
    ky = transverse_wavevectors[..., 1]
    return jnp.block(
        [
            [
                utils.diag(ky / permittivity[..., jnp.newaxis] * ky),
                utils.diag(-ky / permittivity[..., jnp.newaxis] * kx),
            ],
            [
                utils.diag(-kx / permittivity[..., jnp.newaxis] * ky),
                utils.diag(kx / permittivity[..., jnp.newaxis] * kx),
            ],
        ]
    )


def script_k_matrix_patterned(
    z_permittivity_matrix: jnp.ndarray,
    transverse_wavevectors: jnp.ndarray,
) -> jnp.ndarray:
    """Returns the patterned-layer script-k matrix from eq. 19 of [2012 Liu]."""
    dtype = jnp.promote_types(z_permittivity_matrix, transverse_wavevectors)
    kx = transverse_wavevectors[..., 0].astype(dtype)
    ky = transverse_wavevectors[..., 1].astype(dtype)
    z_inv_kx = jnp.linalg.solve(z_permittivity_matrix.astype(dtype), utils.diag(kx))
    z_inv_ky = jnp.linalg.solve(z_permittivity_matrix.astype(dtype), utils.diag(ky))
    return jnp.block(
        [
            [ky[..., :, jnp.newaxis] * z_inv_ky, -ky[..., :, jnp.newaxis] * z_inv_kx],
            [-kx[..., :, jnp.newaxis] * z_inv_ky, kx[..., :, jnp.newaxis] * z_inv_kx],
        ]
    )


def k_matrix_patterned(
    z_permeability_matrix: jnp.ndarray,
    transverse_wavevectors: jnp.ndarray,
) -> jnp.ndarray:
    """Returns the k-matrix for patterned magnetic materials."""
    dtype = jnp.promote_types(z_permeability_matrix, transverse_wavevectors)
    kx = transverse_wavevectors[..., 0].astype(dtype)
    ky = transverse_wavevectors[..., 1].astype(dtype)
    z_inv_kx = jnp.linalg.solve(z_permeability_matrix.astype(dtype), utils.diag(kx))
    z_inv_ky = jnp.linalg.solve(z_permeability_matrix.astype(dtype), utils.diag(ky))
    return jnp.block(
        [
            [kx[..., :, jnp.newaxis] * z_inv_kx, kx[..., :, jnp.newaxis] * z_inv_ky],
            [ky[..., :, jnp.newaxis] * z_inv_kx, ky[..., :, jnp.newaxis] * z_inv_ky],
        ]
    )


# -----------------------------------------------------------------------------
# Functions for computing Fourier convolution matrices for isotropic media.
# -----------------------------------------------------------------------------


def transverse_permittivity_fft(
    permittivity: jnp.ndarray,
    expansion: basis.Expansion,
) -> jnp.ndarray:
    """Computes the `eps` matrix from [2012 Liu] equation 15 using `fft` scheme.

    Args:
        permittivity: The permittivity array, with shape `(..., nx, ny)`.
        expansion: The field expansion to be used.

    Returns:
        The transverse permittivity matrix.
    """
    eps_hat = fft.fourier_convolution_matrix(permittivity, expansion)
    zeros = jnp.zeros_like(eps_hat)
    return jnp.block([[eps_hat, zeros], [zeros, eps_hat]])


def transverse_permittivity_vector(
    permittivity: jnp.ndarray,
    tx: jnp.ndarray,
    ty: jnp.ndarray,
    expansion: basis.Expansion,
) -> jnp.ndarray:
    """Computes transverse permittivity matrix using a vector field methods.

    The transverse permittivity matrix is E2 given in equation 51 of [2012 Liu].

    Args:
        permittivity: The permittivity array, with shape `(..., nx, ny)`.
        tx: The x-component of the tangent vector field.
        ty: The y-component of the tangent vector field.
        expansion: The field expansion to be used.

    Returns:
        The `eps` matrix.
    """
    _transform = functools.partial(fft.fourier_convolution_matrix, expansion=expansion)

    eps_hat = _transform(permittivity)
    zeros = jnp.zeros_like(eps_hat)
    eps_matrix = jnp.block([[eps_hat, zeros], [zeros, eps_hat]])

    eta = 1 / permittivity
    eta_hat = _transform(eta)
    delta_hat = eps_hat - jnp.linalg.inv(eta_hat)
    delta_matrix = jnp.block([[delta_hat, zeros], [zeros, delta_hat]])

    Pxx, Pxy, Pyx, Pyy = _tangent_terms(tx, ty)
    p_matrix = jnp.block(
        [[_transform(Pyy), _transform(Pyx)], [_transform(Pxy), _transform(Pxx)]]
    )
    return eps_matrix - delta_matrix @ p_matrix


def _tangent_terms(
    tx: jnp.ndarray,
    ty: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Returns the `(Pxx, Pxy, Pyx, Pyy)` from [2012 Liu] equation 50.

    Args:
        tx: The x-component of the tangent vector field.
        ty: The y-component of the tangent vector field.

    Returns:
        The terms `(Pxx, Pxy, Pyx, Pyy)`.
    """
    denom = jnp.abs(tx) ** 2 + jnp.abs(ty) ** 2
    denom_safe = jnp.where(jnp.isclose(denom, 0), 1.0, denom)

    # Compute terms from equation 50 of [2012 Liu]. Note that there is an error
    # in equation 48-51 of [2012 Liu], where x- and y- subscripts are swapped for
    # the diagonal blocks. This can be shown by deriving equation 48 from 46, and
    # can also easily be seen by considering the case of a 1D gratng periodic in
    # the x-direction. In this case, the tangent vector field has a y-component
    # only (i.e. tx = 0), and we want to recover the transverse permittivity
    # given by equation 43. This is possible only by swapping the x- and y-
    # subscripts in the diagonal terms of equations 48-51.
    #
    # Note that we also investigated both orderings numerically with the 1D
    # grating example, and found improved convergence with this correction.
    Pyy = jnp.abs(tx) ** 2 / denom_safe
    Pyx = jnp.conj(tx) * ty / denom_safe
    Pxy = tx * jnp.conj(ty) / denom_safe
    Pxx = jnp.abs(ty) ** 2 / denom_safe
    return Pxx, Pxy, Pyx, Pyy


# -----------------------------------------------------------------------------
# Functions for computing Fourier convolution matrices for anisotropic media.
# -----------------------------------------------------------------------------


def transverse_permittivity_fft_anisotropic(
    permittivity_xx: jnp.ndarray,
    permittivity_xy: jnp.ndarray,
    permittivity_yx: jnp.ndarray,
    permittivity_yy: jnp.ndarray,
    expansion: basis.Expansion,
) -> jnp.ndarray:
    """Compute anisotropic transverse permittivity matrix using the fft scheme."""
    _transform = functools.partial(fft.fourier_convolution_matrix, expansion=expansion)
    return jnp.block(
        [
            [_transform(permittivity_yy), _transform(-permittivity_yx)],
            [_transform(-permittivity_xy), _transform(permittivity_xx)],
        ]
    )


def transverse_permeability_fft_anisotropic(
    permeability_xx: jnp.ndarray,
    permeability_xy: jnp.ndarray,
    permeability_yx: jnp.ndarray,
    permeability_yy: jnp.ndarray,
    expansion: basis.Expansion,
) -> jnp.ndarray:
    """Compute anisotropic transverse permeability matrix using the fft scheme."""
    _transform = functools.partial(fft.fourier_convolution_matrix, expansion=expansion)
    return jnp.block(
        [
            [_transform(permeability_xx), _transform(permeability_xy)],
            [_transform(permeability_yx), _transform(permeability_yy)],
        ]
    )


def transverse_permittivity_vector_anisotropic(
    permittivity_xx: jnp.ndarray,
    permittivity_xy: jnp.ndarray,
    permittivity_yx: jnp.ndarray,
    permittivity_yy: jnp.ndarray,
    tx: jnp.ndarray,
    ty: jnp.ndarray,
    expansion: basis.Expansion,
) -> jnp.ndarray:
    """Compute the transverse permittivity matrix with a vector scheme.

    The transverse permittivity matrix E relates the electric and electric displacement
    fields, such that

        [-Dy, Dx]^T = E [-Ey, Ex]^T

    Args:
        permittivity_xx: The xx-component of the permittivity tensor, with
            shape `(..., nx, ny)`.
        permittivity_xy: The xy-component of the permittivity tensor.
        permittivity_yx: The yx-component of the permittivity tensor.
        permittivity_yy: The yy-component of the permittivity tensor.
        tx: The x-component of the tangent vector field.
        ty: The y-component of the tangent vector field.
        expansion: The field expansion to be used.

    Returns:
        The transverse permittivity matrix.
    """
    _transform = functools.partial(fft.fourier_convolution_matrix, expansion=expansion)

    # Define the real-space and Fourier transformed rotation matrices. The rotation
    # matrix is defined such that T [Et, En]^T = [-Ey, Ex].
    (
        rotation_matrix,
        fourier_rotation_matrix,
        fourier_inverse_rotation_matrix,
    ) = _rotation_matrices(
        t00=ty,
        t01=tx.conj(),
        t10=-tx,
        t11=ty.conj(),
        expansion=expansion,
    )

    # Obtain the permittivity tensor for rotated coordinates, i.e. coordinates where
    # unit vectors are tangent and normal to the vector field defined by `(tx, ty)`.
    permittivity_tensor = jnp.block(
        [
            [
                permittivity_yy[..., jnp.newaxis, jnp.newaxis],
                -permittivity_yx[..., jnp.newaxis, jnp.newaxis],
            ],
            [
                -permittivity_xy[..., jnp.newaxis, jnp.newaxis],
                permittivity_xx[..., jnp.newaxis, jnp.newaxis],
            ],
        ]
    )
    assert rotation_matrix.shape[-2:] == permittivity_tensor.shape[-2:] == (2, 2)
    rotated_permittivity_tensor = jnp.linalg.solve(
        rotation_matrix, permittivity_tensor @ rotation_matrix
    )

    # The Fourier permittivity matrix is the central matrix in equation 45 of
    # [2012 Liu]. The top left block relates the tangential E and D fields, while
    # the bottom right block relates the normal E and D fields. Consequently, these
    # use Laurent's rule and the inverse rule, respectively. Note that the
    # non-diagonal blocks also use Laurent's rule, as these tensor elements may be
    # zero, making the inverse rule problematic.
    fourier_permittivity_matrix = jnp.block(
        [
            [
                _transform(rotated_permittivity_tensor[..., 0, 0]),
                _transform(rotated_permittivity_tensor[..., 0, 1]),
            ],
            [
                _transform(rotated_permittivity_tensor[..., 1, 0]),
                jnp.linalg.inv(_transform(1 / rotated_permittivity_tensor[..., 1, 1])),
            ],
        ]
    )
    return (
        fourier_rotation_matrix
        @ fourier_permittivity_matrix
        @ fourier_inverse_rotation_matrix
    )


def transverse_permeability_vector_anisotropic(
    permeability_xx: jnp.ndarray,
    permeability_xy: jnp.ndarray,
    permeability_yx: jnp.ndarray,
    permeability_yy: jnp.ndarray,
    tx: jnp.ndarray,
    ty: jnp.ndarray,
    expansion: basis.Expansion,
) -> jnp.ndarray:
    """Compute the transverse permeability matrix with a vector scheme.

    The transverse permeability matrix M relates the magnetic and magnetic flux
    density fields, such that

        [Bx, Bx]^T = M [Hx, Hy]^T

    Important differences from the `_transverse_permittivity_vector_anisotropic`
    function result from the different definitions of E and M matrices.

    Args:
        permeability_xx: The xx-component of the permeability tensor, with
            shape `(..., nx, ny)`.
        permeability_xy: The xy-component of the permeability tensor.
        permeability_yx: The yx-component of the permeability tensor.
        permeability_yy: The yy-component of the permeability tensor.
        tx: The x-component of the tangent vector field.
        ty: The y-component of the tangent vector field.
        expansion: The field expansion to be used.

    Returns:
        The transverse permeability matrix.
    """
    _transform = functools.partial(fft.fourier_convolution_matrix, expansion=expansion)

    # Define the real-space and Fourier transformed rotation matrices. The rotation
    # matrix is defined such that T [Ht, Hn]^T = [Hx, Hy], and differs from that
    # used for the transverse permittivity calculation.
    (
        rotation_matrix,
        fourier_rotation_matrix,
        fourier_inverse_rotation_matrix,
    ) = _rotation_matrices(
        t00=tx,
        t01=-ty.conj(),
        t10=ty,
        t11=tx.conj(),
        expansion=expansion,
    )

    # Obtain the permeability tensor for rotated coordinates, i.e. coordinates where
    # unit vectors are tangent and normal to the vector field defined by `(tx, ty)`.
    permeability_tensor = jnp.block(
        [
            [
                permeability_xx[..., jnp.newaxis, jnp.newaxis],
                permeability_xy[..., jnp.newaxis, jnp.newaxis],
            ],
            [
                permeability_yx[..., jnp.newaxis, jnp.newaxis],
                permeability_yy[..., jnp.newaxis, jnp.newaxis],
            ],
        ]
    )
    assert rotation_matrix.shape[-2:] == permeability_tensor.shape[-2:] == (2, 2)
    rotated_permeability_tensor = jnp.linalg.solve(
        rotation_matrix, permeability_tensor @ rotation_matrix
    )

    # The Fourier permeability matrix is analogous to the central matrix in equation 45
    # of [2012 Liu], in that its top-left block relates the tangential magnetic and
    # magnetic flux density fields, and the bottom-right block relates the normal
    # fields. Consequently, these use Laurent's rule and the inverse rule, respectively.
    # Note that the non-diagonal blocks also use Laurent's rule, as these tensor
    # elements may be zero, making the inverse rule problematic.
    fourier_permeability_matrix = jnp.block(
        [
            [
                _transform(rotated_permeability_tensor[..., 0, 0]),  # tt
                _transform(rotated_permeability_tensor[..., 0, 1]),  # tn
            ],
            [
                _transform(rotated_permeability_tensor[..., 1, 0]),  # nt
                jnp.linalg.inv(
                    _transform(1 / rotated_permeability_tensor[..., 1, 1])
                ),  # nn
            ],
        ]
    )
    return (
        fourier_rotation_matrix
        @ fourier_permeability_matrix
        @ fourier_inverse_rotation_matrix
    )


def _rotation_matrices(
    t00: jnp.ndarray,
    t01: jnp.ndarray,
    t10: jnp.ndarray,
    t11: jnp.ndarray,
    expansion: basis.Expansion,
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Generate real-space and Fourier transformed rotation matrices."""
    rotation_matrix = jnp.block(
        [
            [
                t00[..., jnp.newaxis, jnp.newaxis],
                t01[..., jnp.newaxis, jnp.newaxis],
            ],
            [
                t10[..., jnp.newaxis, jnp.newaxis],
                t11[..., jnp.newaxis, jnp.newaxis],
            ],
        ]
    )

    # Blockwise Fourier transform of the rotation matrix.
    _transform = functools.partial(fft.fourier_convolution_matrix, expansion=expansion)
    fourier_rotation_matrix = jnp.block(
        [
            [_transform(t00), _transform(t01)],
            [_transform(t10), _transform(t11)],
        ]
    )

    # Blockwise Fourier transform of the inverse rotation matrix.
    denom = jnp.abs(t00 * t11 - t10 * t01)
    denom_safe = jnp.where(jnp.isclose(denom, 0), 1.0, denom)
    fourier_inverse_rotation_matrix = jnp.block(
        [
            [_transform(t11 / denom_safe), _transform(-t01 / denom_safe)],
            [_transform(-t10 / denom_safe), _transform(t00 / denom_safe)],
        ]
    )
    return rotation_matrix, fourier_rotation_matrix, fourier_inverse_rotation_matrix



"""Functions related to sources in the FMM algorithm.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

from typing import Optional, Tuple

import jax.numpy as jnp

from fmmax import basis, fft, fields, fmm, scattering, utils


def amplitudes_for_fields(
    ex: jnp.ndarray,
    ey: jnp.ndarray,
    hx: jnp.ndarray,
    hy: jnp.ndarray,
    layer_solve_result: fmm.LayerSolveResult,
    brillouin_grid_axes: Optional[Tuple[int, int]] = None,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Computes the amplitudes for fields.

    The fields are expected to be given for a supercell containing one or more unit
    cells, i.e. for a calculation with in-plane wavevector generated by
    `basis.brillouin_zone_in_plane_wavevector`. The shape of the supercell is
    inferred from the batch shape of the `layer_solve_result` and the specified
    `brillouin_grid_axes`.

    Args:
        ex: The x-oriented electric field on the real-space grid, with a trailing
            batch dimension.
        ey: The y-oriented electric field on the real-space grid, with shape
            `(..., nx, ny, num_fields)`.
        hx: The x-oriented magnetic field on the real-space grid.
        hy: The y-oriented magnetic field on the real-space grid.
        layer_solve_result: The eigensolve result for the layer containing the fields.
        brillouin_grid_axes: The batch axes corresponding to the Brillouin zone grid.

    Returns:
        The forward and backward wave amplitudes.
    """

    if not (ex.shape == ey.shape == hx.shape == hy.shape) or (ex.ndim < 3):
        raise ValueError(
            f"All fields must have rank of at least 3 with matching shape, but got "
            f"shapes of {ex.shape}, {ey.shape}, {hx.shape}, and {hy.shape}."
        )

    if not utils.batch_compatible_shapes(
        layer_solve_result.batch_shape,
        ex.shape[:-3],
        ey.shape[:-3],
        hx.shape[:-3],
        hy.shape[:-3],
    ):
        raise ValueError(
            f"Fields must be batch-compatible with `layer_solve_result`, but got "
            f"shapes of {ex.shape}, {ey.shape}, {hx.shape}, and {hy.shape} with "
            f"`layer_solve_result` batch shape {layer_solve_result.batch_shape}."
        )

    if brillouin_grid_axes is None:
        brillouin_grid_shape = (1, 1)
    else:
        absolute_brillouin_grid_axes = utils.absolute_axes(brillouin_grid_axes, ex.ndim)
        brillouin_grid_shape = (
            layer_solve_result.batch_shape[absolute_brillouin_grid_axes[0]],
            layer_solve_result.batch_shape[absolute_brillouin_grid_axes[1]],
        )

    if (
        ex.shape[-3] % brillouin_grid_shape[0] != 0
        or ex.shape[-2] % brillouin_grid_shape[1] != 0
    ):
        raise ValueError(
            f"Field shapes must be evenly divisible by the Brillouin grid shape, but "
            f"got shape {ex.shape} when inferred Brillouin grid shape is "
            f"{brillouin_grid_shape}."
        )

    unit_cell_shape = (
        ex.shape[-3] // brillouin_grid_shape[0],
        ex.shape[-2] // brillouin_grid_shape[1],
    )

    x, y = basis.unit_cell_coordinates(
        primitive_lattice_vectors=layer_solve_result.primitive_lattice_vectors,
        shape=unit_cell_shape,
        num_unit_cells=brillouin_grid_shape,
    )
    kx = layer_solve_result.in_plane_wavevector[..., 0, jnp.newaxis, jnp.newaxis]
    ky = layer_solve_result.in_plane_wavevector[..., 1, jnp.newaxis, jnp.newaxis]
    phase = jnp.exp(1j * (kx * x + ky * y))[..., jnp.newaxis]

    def _process_field(field: jnp.ndarray) -> jnp.ndarray:
        field /= phase
        field_split = jnp.asarray(
            tuple(
                jnp.split(s, indices_or_sections=brillouin_grid_shape[1], axis=-2)
                for s in jnp.split(
                    field, indices_or_sections=brillouin_grid_shape[0], axis=-3
                )
            )
        )
        field_fft = fft.fft(
            field_split, expansion=layer_solve_result.expansion, axes=(-3, -2)
        )
        return jnp.sum(field_fft, axis=(0, 1))

    ex = _process_field(ex)
    ey = _process_field(ey)
    hx = _process_field(hx)
    hy = _process_field(hy)

    stacked_fields = jnp.concatenate([-ey, ex, hx, hy], axis=-2)
    matrix = fields.field_conversion_matrix(layer_solve_result)
    amplitudes = jnp.linalg.solve(matrix, stacked_fields)

    fwd_amplitude, bwd_amplitude = jnp.split(amplitudes, indices_or_sections=2, axis=-2)
    return fwd_amplitude, bwd_amplitude


# -----------------------------------------------------------------------------
# Functions related to internal sources.
# -----------------------------------------------------------------------------


def dirac_delta_source(
    location: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    expansion: basis.Expansion,
) -> jnp.ndarray:
    """Returns the coefficients for a delta source at the specified location.

    This function is appropriate for creating sources to be used with
    `amplitudes_for_source`.

    Args:
        location: The location of the source, with shape `(num_sources, 2)` and
            the trailing axis giving the x and y location. By convention, the
            center of the unit cell is at `(0, 0)`.
        in_plane_wavevector: The in-plane wavevevector for the calculation, which
            gives the offset of the plane wave decomposition. Has shape `(..., 2)`
            with possible batch dimensions.
        primitive_lattice_vectors: The primitive lattice vectors of the unit cell.
        expansion: The Fourier expansion used for the calculation.

    Returns:
        The coefficients, with the shape `(..., expansion.num_terms, num_sources)`.
    """
    if location.ndim != 2 or location.shape[-1] != 2:
        raise ValueError(
            f"`location` must be rank-2 with a trailing axis size of 2, but got "
            f"shape {location.shape}"
        )
    if in_plane_wavevector.shape[-1] != 2:
        raise ValueError(
            f"`in_plane_wavevector` must have a trailing axis size of 2, but got "
            f"shape {in_plane_wavevector.shape}"
        )

    transverse_wavevectors = basis.transverse_wavevectors(
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
    )
    kx = transverse_wavevectors[..., 0][..., jnp.newaxis]
    ky = transverse_wavevectors[..., 1][..., jnp.newaxis]
    x = location[:, 0]
    y = location[:, 1]
    return jnp.exp(-1j * (kx * x + ky * y))


def gaussian_source(
    fwhm: jnp.ndarray,
    location: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    expansion: basis.Expansion,
) -> jnp.ndarray:
    """Returns the coefficients for a Gaussian source at the specified location.

    This function is appropriate for creating sources to be used with
    `amplitudes_for_source`.

    Args:
        fwhm: The full-width at half-maximum for the Gaussian source.
        location: The location of the source, with shape `(num_sources, 2)` and
            the trailing axis giving the x and y location. By convention, the
            center of the unit cell is at `(0, 0)`.
        in_plane_wavevector: The in-plane wavevevector for the calculation, which
            gives the offset of the plane wave decomposition. Has shape `(..., 2)`
            with possible batch dimensions.
        primitive_lattice_vectors: The primitive lattice vectors of the unit cell.
        expansion: The Fourier expansion used for the calculation.

    Returns:
        The coefficients, with the shape `(..., expansion.num_terms, num_sources)`.
    """
    if location.ndim != 2 or location.shape[-1] != 2:
        raise ValueError(
            f"`location` must be rank-2 with a trailing axis size of 2, but got "
            f"shape {location.shape}"
        )
    if in_plane_wavevector.shape[-1] != 2:
        raise ValueError(
            f"`in_plane_wavevector` must have a trailing axis size of 2, but got "
            f"shape {in_plane_wavevector.shape}"
        )

    transverse_wavevectors = basis.transverse_wavevectors(
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
    )
    kx = transverse_wavevectors[..., 0][..., jnp.newaxis]
    ky = transverse_wavevectors[..., 1][..., jnp.newaxis]
    x = location[:, 0]
    y = location[:, 1]
    coeffs = jnp.exp(-1j * (kx * x + ky * y))

    sigma = fwhm / (2 * jnp.sqrt(2 * jnp.log(2)))
    return coeffs * jnp.exp(-(sigma**2) * (kx**2 + ky**2) / 2)


def amplitudes_for_source(
    jx: jnp.ndarray,
    jy: jnp.ndarray,
    jz: jnp.ndarray,
    s_matrix_before_source: scattering.ScatteringMatrix,
    s_matrix_after_source: scattering.ScatteringMatrix,
) -> Tuple[
    jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray
]:
    """Computes wave amplitudes resulting from an internal source.

    The configuration of the calculation is depicted below. A source is located at
    the interface of two layers. The layer stacks before and after the source are
    named as such. The function computes the amplitudes of forward-going and
    backward-going waves at various locations within the stack, also depicted below.

                                source _____
                                            |
             __before_______________________V___after_____________________

            |             |     |           |           |     |           |
            |   layer 0   | ... |  layer i  |   layer   | ... |  layer N  |
            | start layer |     |           |   i + 1   |     | end layer |
            |             |     |           |           |     |           |

                                 -> a_i      -> a_i+1          -> a_N
                    b_0 <-            b_i <-    b_i+1 <-

    Args:
        jx: The x-oriented dipole amplitude; must be at least rank-3 with a
            trailing batch dimension.
        jy: The y-oriented dipole amplitude, with shape matching `jx`.
        jz: The z-oriented dipole amplitude, with shape matching `jx`.
        s_matrix_before_source: The scattering matrix for the layer substack
            before the source, having no overlap with the after-source substack.
            Scattering matrix pairs returned by `scattering.stack_s_matrices_interior`
            may not be directly used.
        s_matrix_after_source: The scattering matrix for the layer substack after
            the source.

    Returns:
        The wave amplitudes:
            backward_amplitude_0_end: The backward-going wave amplitude at the end
                of the first layer.
            forward_amplitude_before_start: The forward-going wave amplitude at the
                start of the layer preceding the source.
            backward_amplitude_before_end: The backward-going wave amplitude at the
                end of the layer preceding the source, i.e. just before the source.
            forward_amplitude_after_start: The forward-going wave amplitude at the
                start of the layer following the source, i.e. just after the source.
            backward_amplitude_after_end: The backward-going wave amplitude at the
                end of the layer following the source.
            forward_amplitude_N_start: The forward-going wave amplitude at the start
                of the final layer.
    """
    pol = polarization_terms(
        jx, jy, jz, layer_solve_result=s_matrix_before_source.end_layer_solve_result
    )
    matrix = emission_matrix(
        s_matrix_before_source=s_matrix_before_source,
        s_matrix_after_source=s_matrix_after_source,
    )

    # The matrix solve directly yields the forward-going wave amplitudes after the
    # source, and the backward-going wave amplitudes before the source.
    ab = jnp.linalg.solve(matrix, pol)
    fwd_amplitude_after_start, bwd_amplitude_before_end = jnp.split(ab, 2, axis=-2)

    # Other wave amplitudes are computed from the scattering matrices.
    bwd_amplitude_0_end = s_matrix_before_source.s22 @ bwd_amplitude_before_end
    fwd_amplitude_before_start = s_matrix_before_source.s12 @ bwd_amplitude_before_end
    bwd_amplitude_after_end = s_matrix_after_source.s21 @ fwd_amplitude_after_start
    fwd_amplitude_N_start = s_matrix_after_source.s11 @ fwd_amplitude_after_start

    return (
        bwd_amplitude_0_end,
        fwd_amplitude_before_start,
        bwd_amplitude_before_end,
        fwd_amplitude_after_start,
        bwd_amplitude_after_end,
        fwd_amplitude_N_start,
    )


def polarization_terms(
    jx: jnp.ndarray,
    jy: jnp.ndarray,
    jz: jnp.ndarray,
    layer_solve_result: fmm.LayerSolveResult,
) -> jnp.ndarray:
    """Computes the polarization terms for currents on the real-space grid.

    The polarization terms are discussed in section 7 of [1999 Whittaker].

    Args:
        jx: The Fourier-transformed x-oriented dipole amplitude, with a
            trailing batch dimension.
        jy: The Fourier-transformed y-oriented dipole amplitude.
        jz: The Fourier-transformed z-oriented dipole amplitude.
        layer_solve_result: The results of the layer eigensolve.

    Returns:
        The polarization vector containing parallel and z-oriented terms.
    """
    if not (jx.shape == jy.shape == jz.shape) or jx.ndim < 2:
        raise ValueError(
            f"`jx`, `jy`, and `jz` must have matching shape and be at least "
            f"rank-2, but got shapes {jx.shape, jy.shape, jz.shape}."
        )

    # Add dimensions as needed to ensures shapes are compatible.
    eta_matrix = layer_solve_result.inverse_z_permittivity_matrix
    leading_ndim = max(jx.ndim - 3, eta_matrix.ndim - 2)
    eta_matrix = utils.atleast_nd(eta_matrix, n=(leading_ndim + 2))
    jx = utils.atleast_nd(jx, n=(leading_ndim + 2))
    jy = utils.atleast_nd(jy, n=(leading_ndim + 2))
    jz = utils.atleast_nd(jz, n=(leading_ndim + 2))

    p_parallel = jnp.concatenate([jy, -jx], axis=-2)

    transverse_wavevectors = basis.transverse_wavevectors(
        in_plane_wavevector=layer_solve_result.in_plane_wavevector,
        primitive_lattice_vectors=layer_solve_result.primitive_lattice_vectors,
        expansion=layer_solve_result.expansion,
    )
    kx = transverse_wavevectors[..., 0, jnp.newaxis]
    ky = transverse_wavevectors[..., 1, jnp.newaxis]

    # Use the `eta_matrix`, the Fourier convolution matrix for the inverse of
    # the permittivity rather than inverting the Fourier convolution matrix for
    # permittivity itself. The improved convergence due to this choice is
    # discussed in [1999 Whittaker].
    pzy = ky * eta_matrix @ jz
    pzx = kx * eta_matrix @ jz
    pz = jnp.concatenate([-pzy, pzx], axis=-2)

    p_parallel = jnp.broadcast_to(p_parallel, pz.shape)
    return jnp.concatenate([pz, p_parallel], axis=-2)


def emission_matrix(
    s_matrix_before_source: scattering.ScatteringMatrix,
    s_matrix_after_source: scattering.ScatteringMatrix,
) -> jnp.ndarray:
    """Returns the emission matrix for a source between two layers.

    Args:
        s_matrix_before_source: The scattering matrix for the layer substack
            before the source, having no overlap with the after-source substack.
            Scattering matrix pairs returned by `scattering.stack_s_matrices_interior`
            may not be directly used.
        s_matrix_after_source: The scattering matrix for the layer substack after
            the source.

    Returns:
        The emission matrix.
    """
    q = s_matrix_before_source.end_layer_solve_result.eigenvalues
    phi = s_matrix_before_source.end_layer_solve_result.eigenvectors
    omega_k = s_matrix_before_source.end_layer_solve_result.omega_script_k_matrix
    layer_thickness = s_matrix_before_source.end_layer_thickness

    next_q = s_matrix_after_source.start_layer_solve_result.eigenvalues
    next_phi = s_matrix_after_source.start_layer_solve_result.eigenvectors
    next_omega_k = s_matrix_after_source.start_layer_solve_result.omega_script_k_matrix
    next_layer_thickness = s_matrix_after_source.start_layer_thickness

    # Compute terms used in the matrix and assemble the matrix.
    fd = jnp.exp(1j * q * layer_thickness)
    next_fd = jnp.exp(1j * next_q * next_layer_thickness)

    fd_s12_before = utils.diag(fd) @ s_matrix_before_source.s12
    fd_s21_after = utils.diag(next_fd) @ s_matrix_after_source.s21

    shape = jnp.broadcast_shapes(q.shape, next_q.shape)
    eye = utils.diag(jnp.ones(shape, dtype=q.dtype))

    # Equation 7.9 from [1999 Whittaker].
    matrix = jnp.block(
        [
            [
                next_omega_k @ next_phi @ utils.diag(1 / next_q) @ (eye - fd_s21_after),
                omega_k @ phi @ utils.diag(1 / q) @ (eye - fd_s12_before),
            ],
            [
                next_phi @ (eye + fd_s21_after),
                -phi @ (eye + fd_s12_before),
            ],
        ]
    )

    return matrix



"""Defines several utility functions.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

from typing import Tuple

import jax
import jax.numpy as jnp

# The `jeig` package offers several jax-wrapped implementations of eigendecomposition,
# some of which have performance benefits. However, since `jeig` has a dependency on
# pytorch, we make its use optional. If `jeig` is not available, we fall back on a
# pure-jax implementation of the eigendecomposition.
try:
    import jeig

    _JEIG_AVAILABLE = True
except ModuleNotFoundError:
    _JEIG_AVAILABLE = False


EIG_EPS_RELATIVE = 1e-12
EIG_EPS_MINIMUM = 1e-24


def diag(x: jnp.ndarray) -> jnp.ndarray:
    """A batch-compatible version of `numpy.diag`."""
    shape = x.shape + (x.shape[-1],)
    y = jnp.zeros(shape, x.dtype)
    i = jnp.arange(x.shape[-1])
    return y.at[..., i, i].set(x)


def angular_frequency_for_wavelength(wavelength: jnp.ndarray) -> jnp.ndarray:
    """Returns the angular frequency for the specified wavelength."""
    return 2 * jnp.pi / wavelength  # Since by our convention c == 1.


def matrix_adjoint(x: jnp.ndarray) -> jnp.ndarray:
    """Computes the adjoint for a batch of matrices."""
    axes = tuple(range(x.ndim - 2)) + (x.ndim - 1, x.ndim - 2)
    return jnp.conj(jnp.transpose(x, axes=axes))


def batch_compatible_shapes(*shapes: Tuple[int, ...]) -> bool:
    """Returns `True` if all the shapes are batch-compatible."""
    max_dims = max([len(s) for s in shapes])
    shapes = tuple([(1,) * (max_dims - len(s)) + s for s in shapes])
    max_shape = [max(dim_shapes) for dim_shapes in zip(*shapes)]
    for shape in shapes:
        if any([a not in (1, b) for a, b in zip(shape, max_shape)]):
            return False
    return True


def atleast_nd(x: jnp.ndarray, n: int) -> jnp.ndarray:
    """Adds leading dimensions to `x`, ensuring that it is at least n-dimensional."""
    dims_to_add = tuple(range(max(0, n - x.ndim)))
    return jnp.expand_dims(x, axis=dims_to_add)


def absolute_axes(axes: Tuple[int, ...], ndim: int) -> Tuple[int, ...]:
    """Returns the absolute axes for given relative axes and array dimensionality."""
    if not all(a in list(range(-ndim, ndim)) for a in axes):
        raise ValueError(
            f"All elements of `axes` must be in the range ({ndim}, {ndim - 1}) "
            f"but got {axes}."
        )
    absolute_axes = tuple([d % ndim for d in axes])
    if len(absolute_axes) != len(set(absolute_axes)):
        raise ValueError(
            f"Found duplicates in `axes`; computed absolute axes are {absolute_axes}."
        )
    return absolute_axes


def interpolate_permittivity(
    permittivity_solid: jnp.ndarray,
    permittivity_void: jnp.ndarray,
    density: jnp.ndarray,
) -> jnp.ndarray:
    """Interpolates the permittivity with a scheme that avoids zero crossings.

    The interpolation uses the scheme introduced in [2019 Christiansen], which avoids
    zero crossings that can occur with metals or lossy materials having a negative
    real component of the permittivity. https://doi.org/10.1016/j.cma.2018.08.034

    Args:
        permittivity_solid: The permittivity of solid regions.
        permittivity_void: The permittivity of void regions.
        density: The density, specifying which locations are solid and which are void.

    Returns:
        The interpolated permittivity.
    """
    n_solid = jnp.real(jnp.sqrt(permittivity_solid))
    k_solid = jnp.imag(jnp.sqrt(permittivity_solid))
    n_void = jnp.real(jnp.sqrt(permittivity_void))
    k_void = jnp.imag(jnp.sqrt(permittivity_void))
    n = density * n_solid + (1 - density) * n_void
    k = density * k_solid + (1 - density) * k_void
    return (n + 1j * k) ** 2


# -----------------------------------------------------------------------------
# Functions related to a generalized eigensolve with custom vjp rule.
# -----------------------------------------------------------------------------


@jax.custom_vjp
def eig(
    matrix: jnp.ndarray,
    eps_relative: float = EIG_EPS_RELATIVE,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Wraps `jnp.linalg.eig` in a jit-compatible, differentiable manner.

    The custom vjp allows gradients with resepct to the eigenvectors, unlike the
    standard jax implementation of `eig`. We use an expression for the gradient
    given in [2019 Boeddeker] along with a regularization scheme that applies
    a Lorentzian broadening to a term containing the inverse difference of
    eigenvalues. The broadening is related to the maximum magnitude of the
    eigenvalues.

    [2019 Boeddeker] https://arxiv.org/abs/1701.00392

    Args:
        matrix: The matrix for which eigenvalues and eigenvectors are sought.
        eps_relative: Parameter which determines the degree of broadening.

    Returns:
        The eigenvalues and eigenvectors.
    """
    del eps_relative
    return _eig(matrix)


def _eig(matrix: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Eigendecomposition using `jeig` if available, and `jax.lax.linalg.eig` if not."""
    if _JEIG_AVAILABLE:
        return jeig.eig(matrix)
    else:
        eigenvalues, eigenvectors = jax.lax.linalg.eig(
            matrix, compute_left_eigenvectors=False, use_magma=False
        )
        return eigenvalues, eigenvectors


def _eig_fwd(
    matrix: jnp.ndarray,
    eps_relative: float,
) -> Tuple[Tuple[jnp.ndarray, jnp.ndarray], Tuple[jnp.ndarray, jnp.ndarray, float]]:
    """Implements the forward calculation for `eig`."""
    eigenvalues, eigenvectors = _eig(matrix)
    return (eigenvalues, eigenvectors), (eigenvalues, eigenvectors, eps_relative)


def _eig_bwd(
    res: Tuple[jnp.ndarray, jnp.ndarray, float],
    grads: Tuple[jnp.ndarray, jnp.ndarray],
) -> Tuple[jnp.ndarray, None]:
    """Implements the backward calculation for `eig`."""
    eigenvalues, eigenvectors, eps_relative = res
    grad_eigenvalues, grad_eigenvectors = grads

    # The expression for gradient of matrix eigenvectors with respect to matrix values
    # contains a difference between eigenvalues in the denominator. This causes
    # numerical problems in situations where eigenvalues are degenerate or nearly
    # degenerate. This problem is addressed in a few different ways in existing rcwa
    # codes. Defining `delta = eigval_i - eigval_j`,
    #
    # - torcwa uses Lorentzian broadening of the form
    #      1 / delta -> delta.conj / (abs(delta)**2 + eps)
    #   with an eps value of 1e-10.
    #   (https://github.com/kch3782/torcwa/blob/main/torcwa/torch_eig.py#L29)
    #
    # - tf_rcwa uses an expression similar to torcwa, but with apparently different
    #   conjugation. Their value of eps is 1e-6. (We tested their implementation and
    #   found it fails some of our gradient validation tests.)
    #   (https://github.com/scolburn54/rcwa_tf/blob/master/src/tensor_utils.py#L89)
    #
    # - grcwa uses `1 / delta -> 1 / (delta + eps)` with an eps value of 1e-10.
    #   (https://github.com/weiliangjinca/grcwa/blob/master/grcwa/primitives.py#L44)
    #
    # Also, from the wider literature:
    #
    # - Liao et al. (https://arxiv.org/pdf/1903.09650 sec III.A.1) uses Lorentzian
    #   broadening with an eps value of 1e-12.
    #
    # One issue with Lorentzian broadening and a fixed eps is the independence from
    # matrix scale. While a given eps may be appropriate for a matrix, it would be
    # inappropriate for the same matrix scaled by e.g. 1e-6. (The gradient should
    # be identical, except scaled by 1 / 1e-6.)
    #
    # Therefore, we use Lorentzian broadening similar to torcwa, but with an eps
    # value that is computed in a way that considers the eigenvalue range.
    eigenvalues_i = eigenvalues[..., jnp.newaxis, :]
    eigenvalues_j = eigenvalues[..., :, jnp.newaxis]
    delta_eig = eigenvalues_i - eigenvalues_j
    eig_range_sq = jnp.amax(jnp.abs(delta_eig) ** 2, axis=(-2, -1), keepdims=True)
    eps = jnp.maximum(eps_relative * eig_range_sq, EIG_EPS_MINIMUM)
    f_broadened = delta_eig.conj() / (jnp.abs(delta_eig) ** 2 + eps)

    # Manually set the diagonal elements to zero, as we do not use broadening here.
    i = jnp.arange(f_broadened.shape[-1])
    f_broadened = f_broadened.at[..., i, i].set(0)

    # By jax convention, gradients are with respect to the complex parameters, not with
    # respect to their conjugates. Take the conjugates.
    grad_eigenvalues_conj = jnp.conj(grad_eigenvalues)
    grad_eigenvectors_conj = jnp.conj(grad_eigenvectors)

    eigenvectors_H = matrix_adjoint(eigenvectors)
    dim = eigenvalues.shape[-1]
    eye_mask = jnp.eye(dim, dtype=bool)
    eye_mask = eye_mask.reshape((1,) * (eigenvalues.ndim - 1) + (dim, dim))

    # Then, the gradient is found by equation 4.77 of [2019 Boeddeker].
    rhs = (
        diag(grad_eigenvalues_conj)
        + jnp.conj(f_broadened) * (eigenvectors_H @ grad_eigenvectors_conj)
        - jnp.conj(f_broadened)
        * (eigenvectors_H @ eigenvectors)
        @ jnp.where(eye_mask, jnp.real(eigenvectors_H @ grad_eigenvectors_conj), 0.0)
    ) @ eigenvectors_H
    grad_matrix = jnp.linalg.solve(eigenvectors_H, rhs)

    # Take the conjugate of the gradient, reverting to the jax convention
    # where gradients are with respect to complex parameters.
    grad_matrix = jnp.conj(grad_matrix)

    # Return `grad_matrix`, and `None` for the gradient with respect to `eps`.
    return grad_matrix, None


eig.defvjp(_eig_fwd, _eig_bwd)



"""Functions related transforming to and from the Fourier basis.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

from typing import Tuple

import jax
import jax.numpy as jnp
import numpy as onp

from fmmax import basis, utils


def fourier_convolution_matrix(
    x: jnp.ndarray,
    expansion: basis.Expansion,
) -> jnp.ndarray:
    """Computes the Fourier convolution matrix for `x` and `basis_coefficients`.

    The Fourier convolution matrix at location `(i, j)` gives the Fourier
    coefficient associated with the lattice vector obtained by subtracting the
    `j`th reciprocal lattice vector from the `i`th reciprocal lattice basis.
    See equation 8 from [2012 Liu].

    Args:
        x: The array for which the Fourier coefficients are sought.
        expansion: The field expansion to be used.

    Returns:
        The coefficients, with shape `(num_vectors, num_vectors)`.
    """
    _validate_shape_for_expansion(x.shape, expansion)

    x_fft = jnp.fft.fft2(x)
    x_fft /= jnp.prod(jnp.asarray(x.shape[-2:])).astype(x.dtype)
    idx = _standard_toeplitz_indices(expansion)
    return x_fft[..., idx[..., 0], idx[..., 1]]


def _standard_toeplitz_indices(expansion: basis.Expansion) -> jnp.ndarray:
    """Computes the indices for a standard Toeplitz matrix for `basis_coefficients`.

    Args:
        expansion: The field expansion to be used.

    Returns:
        The indices, with shape `(num, num, 2)`.
    """
    i, j = jnp.meshgrid(
        jnp.arange(expansion.num_terms),
        jnp.arange(expansion.num_terms),
        indexing="ij",
    )
    basis_coefficients = jnp.asarray(expansion.basis_coefficients)
    idx = basis_coefficients[i, :] - basis_coefficients[j, :]
    return idx


def fft(
    x: jnp.ndarray,
    expansion: basis.Expansion,
    axes: Tuple[int, int] = (-2, -1),
) -> jnp.ndarray:
    """Returns the 2D Fourier transform of `x`.

    Args:
        x: The array to be transformed.
        expansion: The field expansion to be used.
        axes: The axes to be transformed, with default being `(-2, -1)`, the final axes.

    Returns:
        The transformed `x`.
    """
    axes: Tuple[int, int] = utils.absolute_axes(axes, x.ndim)  # type: ignore[no-redef]
    _validate_shape_for_expansion(tuple([x.shape[ax] for ax in axes]), expansion)

    x_fft = jnp.fft.fft2(x, axes=axes, norm="forward")

    leading_dims = len(x.shape[: axes[0]])
    trailing_dims = len(x.shape[axes[1] + 1 :])
    slices = (
        [slice(None)] * leading_dims
        + [expansion.basis_coefficients[:, 0], expansion.basis_coefficients[:, 1]]
        + [slice(None)] * trailing_dims
    )
    return x_fft[tuple(slices)]


def ifft(
    y: jnp.ndarray,
    expansion: basis.Expansion,
    shape: Tuple[int, int],
    axis: int = -1,
) -> jnp.ndarray:
    """Returns the 2D inverse Fourier transform of `x`.

    Args:
        y: The array to be transformed.
        expansion: The field expansion to be used.
        shape: The desired shape of the output array.
        axis: The axis containing the Fourier coefficients. Default is `-1`, the
            final axis.

    Returns:
        The inverse transformed `x`.
    """
    (axis,) = utils.absolute_axes((axis,), y.ndim)
    assert y.shape[axis] == expansion.basis_coefficients.shape[-2]
    x_shape = y.shape[:axis] + shape + y.shape[axis + 1 :]
    assert len(x_shape) == y.ndim + 1

    _validate_shape_for_expansion(shape, expansion)

    leading_dims = len(y.shape[:axis])
    trailing_dims = len(y.shape[axis + 1 :])
    slices = (
        [slice(None)] * leading_dims
        + [expansion.basis_coefficients[:, 0], expansion.basis_coefficients[:, 1]]
        + [slice(None)] * trailing_dims
    )

    x = jnp.zeros(x_shape, y.dtype)
    x = x.at[tuple(slices)].set(y)
    return jnp.fft.ifft2(x, axes=(leading_dims, leading_dims + 1), norm="forward")


def _validate_shape_for_expansion(
    shape: Tuple[int, ...],
    expansion: basis.Expansion,
) -> None:
    """Validates that the shape is sufficient for the provided expansion."""
    min_shape = min_array_shape_for_expansion(expansion)
    if any([d < dmin for d, dmin in zip(shape[-2:], min_shape)]):
        raise ValueError(
            f"`shape` is insufficient for `expansion`, the minimum shape for the "
            f"final two axes is {min_shape} but got shape {shape}."
        )


def min_array_shape_for_expansion(expansion: basis.Expansion) -> Tuple[int, int]:
    """Returns the minimum allowed shape for an array to be expanded."""
    with jax.ensure_compile_time_eval():
        return (
            int(2 * onp.amax(onp.abs(expansion.basis_coefficients[:, 0])) + 1),
            int(2 * onp.amax(onp.abs(expansion.basis_coefficients[:, 1])) + 1),
        )



"""Functions related to beam profiles to be used as sources.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

from typing import Callable, Tuple

import jax.numpy as jnp

# -----------------------------------------------------------------------------
# Functions related to arbitrary reorientation of beams.
# -----------------------------------------------------------------------------

Fields = Tuple[
    Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],  # Ex, Ey, Ez
    Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],  # Hx, Hy, Hz
]


def shifted_rotated_fields(
    field_fn: Callable[[jnp.ndarray, jnp.ndarray, jnp.ndarray], Fields],
    x: jnp.ndarray,
    y: jnp.ndarray,
    z: jnp.ndarray,
    beam_origin_x: jnp.ndarray,
    beam_origin_y: jnp.ndarray,
    beam_origin_z: jnp.ndarray,
    polar_angle: jnp.ndarray,
    azimuthal_angle: jnp.ndarray,
    polarization_angle: jnp.ndarray,
) -> Fields:
    """Computes the fields on a rotated coordinate system.

    Given `fields_fn(xf, yf, zf) -> (exf, eyf, ezf), (hxf, hyf, hzf)` which
    returns the fields in the field coordinate system, returns the fields
    at coordinates `(x, y, z)`, which are rotated from `(xf, yf, zf)`, by
    the specified `polar_angle`, `azimuthal_angle`, and `polarization_angle`.

    A beam propagating in the `zf` direction, polarized in the `xf` direction
    will be propagating in the direction specified by `polar_angle` and
    `azimuthal_angle`, with polarization rotated about the propagation
    direction by `polarization_angle`.

    Args:
        field_fn: Function which returns the fields in the field coordinate
            system. The fields should be for a beam propagating in the zf
            direction, i.e. in the z-direction of the beam coordinate system.
        x: x-coordinates of the desired output fields.
        y: y-coordinates of the desired output fields.
        z: z-coordinates of the desired output fields.
        beam_origin_x: The x-origin of the beam coordinate system in the
            `(x, y, z)` unit system.
        beam_origin_y: The y-origin of the beam coordinate system.
        beam_origin_z: The z-origin of the beam coordinate system.
        polar_angle: The rotation angle about the y-axis.
        azimuthal_angle: The rotation angle about the z-axis.
        polarization_angle: The rotation angle about the propagation axis.

    Returns:
        The fields `((ex, ey, ez), (hx, hy, hz))` at the specified coordinates.
    """
    mat = rotation_matrix(polar_angle, azimuthal_angle, polarization_angle)
    mat = jnp.expand_dims(mat, range(x.ndim))

    # Solve for the `(xf, yf, zf)` locations in the field coordinate system
    # which, when rotated as specified, give us the locations `(x, y, z)`.
    assert x.shape == y.shape == z.shape
    coords = jnp.stack([x, y, z], axis=-1)
    rotated_coords = jnp.linalg.solve(mat, coords[..., jnp.newaxis])[..., 0]
    rotated_coords = jnp.split(rotated_coords, 3, axis=-1)
    xf, yf, zf = (jnp.squeeze(r, axis=-1) for r in rotated_coords)

    # Solve for the rotated origin.
    origin = jnp.stack([beam_origin_x, beam_origin_y, beam_origin_z], axis=-1)
    origin = jnp.expand_dims(origin, range(0, mat.ndim - 2))
    rotated_origin = jnp.linalg.solve(mat, origin[..., jnp.newaxis])[..., 0]
    assert rotated_origin.size == 3
    rotated_origin = jnp.split(rotated_origin, 3, axis=-1)
    xf0, yf0, zf0 = (jnp.squeeze(r) for r in rotated_origin)

    # Compute the fields on the rotated, shifted coordinate system.
    (exr, eyr, ezr), (hxr, hyr, hzr) = field_fn(xf - xf0, yf - yf0, zf - zf0)

    rotated_efield = jnp.stack((exr, eyr, ezr), axis=-1)
    rotated_hfield = jnp.stack((hxr, hyr, hzr), axis=-1)

    # Rotate the fields back onto the original coordinate system.
    efield = mat @ rotated_efield[..., jnp.newaxis]
    ex, ey, ez = jnp.split(efield, 3, axis=-2)
    ex = jnp.squeeze(ex, axis=(-2, -1))
    ey = jnp.squeeze(ey, axis=(-2, -1))
    ez = jnp.squeeze(ez, axis=(-2, -1))

    hfield = mat @ rotated_hfield[..., jnp.newaxis]
    hx, hy, hz = jnp.split(hfield, 3, axis=-2)
    hx = jnp.squeeze(hx, axis=(-2, -1))
    hy = jnp.squeeze(hy, axis=(-2, -1))
    hz = jnp.squeeze(hz, axis=(-2, -1))

    return (ex, ey, ez), (hx, hy, hz)


def rotation_matrix(
    polar_angle: jnp.ndarray,
    azimuthal_angle: jnp.ndarray,
    polarization_angle: jnp.ndarray,
) -> jnp.ndarray:
    """Computes a rotation matrix that arbitrarily reorients a field.

    The rotation operations consist of,
      - rotation about the y-axis by `polar_angle`
      - rotation about the z-axis by `azimuthal_angle`
      - rotation about propagation axis by `polarization_angle`, where
        the propagation axis is found by applying the first two rotations
        about the y- and z-axis.

    Args:
        polar_angle: The rotation angle about the y-axis.
        azimuthal_angle: The rotation angle about the z-axis.
        polarization_angle: The rotation angle about the propagation axis.

    Returns:
        The rotation matrix.
    """
    dtype = jnp.promote_types(
        jnp.asarray(polar_angle).dtype,
        jnp.promote_types(
            jnp.asarray(azimuthal_angle).dtype,
            jnp.asarray(polarization_angle).dtype,
        ),
    )
    # Matrix that rotates around the y-axis by `polar_angle`.
    rotation_y_matrix = jnp.asarray(
        [
            [jnp.cos(polar_angle), 0.0, jnp.sin(polar_angle)],
            [0.0, 1.0, 0.0],
            [-jnp.sin(polar_angle), 0.0, jnp.cos(polar_angle)],
        ],
        dtype=dtype,
    )

    # Matrix that rotates around the z-axis by `azimuthal_angle`.
    rotation_z_matrix = jnp.asarray(
        [
            [jnp.cos(azimuthal_angle), -jnp.sin(azimuthal_angle), 0.0],
            [jnp.sin(azimuthal_angle), jnp.cos(azimuthal_angle), 0.0],
            [0.0, 0.0, 1.0],
        ],
        dtype=dtype,
    )

    # Matrix that rotates around the axis defined by the specified polar and
    # azimuthal angle. The unit vector along the axis is `(ux, uy, uz)`.
    ux = jnp.cos(azimuthal_angle) * jnp.sin(polar_angle)
    uy = jnp.sin(azimuthal_angle) * jnp.sin(polar_angle)
    uz = jnp.cos(polar_angle)
    cos_theta_p = jnp.cos(polarization_angle)
    sin_theta_p = jnp.sin(polarization_angle)
    rotation_p_matrix = jnp.asarray(
        [
            [
                cos_theta_p + ux**2 * (1 - cos_theta_p),
                ux * uy * (1 - cos_theta_p) - uz * sin_theta_p,
                ux * uz * (1 - cos_theta_p) + uy * sin_theta_p,
            ],
            [
                uy * ux * (1 - cos_theta_p) + uz * sin_theta_p,
                cos_theta_p + uy**2 * (1 - cos_theta_p),
                uy * uz * (1 - cos_theta_p) - ux * sin_theta_p,
            ],
            [
                uz * ux * (1 - cos_theta_p) - uy * sin_theta_p,
                uz * uy * (1 - cos_theta_p) + ux * sin_theta_p,
                cos_theta_p + uz**2 * (1 - cos_theta_p),
            ],
        ],
        dtype=dtype,
    )
    return rotation_p_matrix @ rotation_z_matrix @ rotation_y_matrix



"""Functions related to tangent vector field generation.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import functools
from typing import Callable, Dict, List, Tuple

import jax
import jax.numpy as jnp

from fmmax import basis, fft, utils

# Absolute tolerance for detecting whether a field is 1D. If the angle of the field at
# every point differs by less than this value from a reference value, the field is 1D.
_ATOL_1D_FIELD_ANGLE = 1e-2


def compute_field_jones_direct(
    arr: jnp.ndarray,
    expansion: basis.Expansion,
    primitive_lattice_vectors: basis.LatticeVectors,
    fourier_loss_weight: float,
    smoothness_loss_weight: float,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Compute tangent vector field using the Jones direct method."""
    return compute_tangent_field(
        arr=arr,
        expansion=expansion,
        primitive_lattice_vectors=primitive_lattice_vectors,
        use_jones_direct=True,
        fourier_loss_weight=fourier_loss_weight,
        smoothness_loss_weight=smoothness_loss_weight,
    )


def compute_field_pol(
    arr: jnp.ndarray,
    expansion: basis.Expansion,
    primitive_lattice_vectors: basis.LatticeVectors,
    fourier_loss_weight: float,
    smoothness_loss_weight: float,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Compute tangent vector field using the Pol method."""
    return compute_tangent_field(
        arr=arr,
        expansion=expansion,
        primitive_lattice_vectors=primitive_lattice_vectors,
        use_jones_direct=False,
        fourier_loss_weight=fourier_loss_weight,
        smoothness_loss_weight=smoothness_loss_weight,
    )


def compute_field_jones(
    arr: jnp.ndarray,
    expansion: basis.Expansion,
    primitive_lattice_vectors: basis.LatticeVectors,
    fourier_loss_weight: float,
    smoothness_loss_weight: float,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Compute tangent vector field using the Jones method."""
    tx, ty = compute_tangent_field(
        arr=arr,
        expansion=expansion,
        primitive_lattice_vectors=primitive_lattice_vectors,
        use_jones_direct=False,
        fourier_loss_weight=fourier_loss_weight,
        smoothness_loss_weight=smoothness_loss_weight,
    )
    jxjy = normalize_jones(jnp.stack([tx, ty], axis=-1))
    jx = jxjy[..., 0]
    jy = jxjy[..., 1]
    return jx, jy


def compute_field_normal(
    arr: jnp.ndarray,
    expansion: basis.Expansion,
    primitive_lattice_vectors: basis.LatticeVectors,
    fourier_loss_weight: float,
    smoothness_loss_weight: float,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Compute tangent vector field using the Normal method."""
    tx, ty = compute_tangent_field(
        arr=arr,
        expansion=expansion,
        primitive_lattice_vectors=primitive_lattice_vectors,
        use_jones_direct=False,
        fourier_loss_weight=fourier_loss_weight,
        smoothness_loss_weight=smoothness_loss_weight,
    )
    txty = normalize_elementwise(jnp.stack([tx, ty], axis=-1))
    tx = txty[..., 0]
    ty = txty[..., 1]
    return tx, ty


# -----------------------------------------------------------------------------
# Underlying functions used to compute all variants of the tangent fields.
# -----------------------------------------------------------------------------


def compute_tangent_field(
    arr: jnp.ndarray,
    expansion: basis.Expansion,
    primitive_lattice_vectors: basis.LatticeVectors,
    use_jones_direct: bool,
    fourier_loss_weight: float,
    smoothness_loss_weight: float,
    steps: int = 1,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Compute the tangent vector field for `arr`.

    The calculation finds the minimum of a quadratic loss function using a single
    Newton iteration. Rather than optimizing the real-space tangent field, the
    Fourier coefficients are directly optimized.

    The tangent field has several properties or invariances:

      - The tangent field is independent of the scale of the unit cell; if the
        unit cell is uniformly scaled (e.g. by changing units from nm to microns),
        the vector field is unchanged.
      - The tangent field for a supercell (containing e.g. 2x2 unit cells) is
        identical to that of a single unit cell, so long as the number of terms in
        the Fourier expansion is increased correspondingly. Note that this means that
        the tangent field depends upon the number of terms in the Fourier expansion.
      - The tangent field is independent of the resolution of the discretized unit
        cell. That is, whether the permittivity distribution is specified with a
        `(100, 100)` or `(200, 200)` shaped array has no impact on the resulting field.
      - The tangent field is independent of the amplitude of the array from which it is
        obtained, e.g. the permittivity contrast.

    Args:
        arr: The array for which the normal vector field is sought.
        expansion: The Fourier expansion for which the field is to be optimized.
        primitive_lattice_vectors: Define the unit cell coordinates.
        use_jones_direct: Specifies whether the complex Jones field is to be sought.
        fourier_loss_weight: Determines the weight of the loss term penalizing
            Fourier terms corresponding to high frequencies. Should be positive.
        smoothness_loss_weight: Determines the weight of the loss term rewarding
            smoothness of the tangent field in real space. Should be positive.
        steps: The number of Newton iterations to carry out. Generally, the default
            single iteration is sufficient to obtain converged fields.

    Returns:
        The normal field, `(tx, ty)`.
    """
    arr = jax.lax.stop_gradient(arr)
    batch_shape = arr.shape[:-2]
    arr = utils.atleast_nd(arr, n=3)
    arr = arr.reshape((-1,) + arr.shape[-2:])

    primitive_lattice_vectors = basis.LatticeVectors(
        u=jnp.broadcast_to(
            primitive_lattice_vectors.u.reshape((-1, 2)),
            (arr.shape[0], 2),
        ),
        v=jnp.broadcast_to(
            primitive_lattice_vectors.v.reshape((-1, 2)),
            (arr.shape[0], 2),
        ),
    )

    field_fn = jax.vmap(
        functools.partial(
            _compute_tangent_field_no_batch,
            use_jones_direct=use_jones_direct,
            fourier_loss_weight=fourier_loss_weight,
            smoothness_loss_weight=smoothness_loss_weight,
            steps=steps,
        ),
        in_axes=(0, None, 0),
    )
    field = field_fn(
        arr,
        expansion,
        primitive_lattice_vectors,
    )
    field = field.reshape(batch_shape + field.shape[-3:])
    tx = field[..., 0]
    ty = field[..., 1]
    return tx, ty


def _compute_tangent_field_no_batch(
    arr: jnp.ndarray,
    expansion: basis.Expansion,
    primitive_lattice_vectors: basis.LatticeVectors,
    use_jones_direct: bool,
    fourier_loss_weight: float,
    smoothness_loss_weight: float,
    steps: int,
) -> jnp.ndarray:
    """Compute the tangent vector field for `arr` with no batch dimensions."""
    assert primitive_lattice_vectors.u.shape == (2,)
    assert arr.ndim == 2
    grid_shape: Tuple[int, int] = arr.shape[-2:]  # type: ignore[assignment]

    # Rescale the weights so that a supercell containing multiple unit cells and
    # having appropriately more terms in the Fourier expansion yields a tangent
    # field identical to that obtained from just a single unit cell.
    fourier_loss_weight /= expansion.num_terms
    smoothness_loss_weight /= expansion.num_terms

    grad = compute_gradient(arr, primitive_lattice_vectors)
    gx = _filter_and_adjust_resolution(grad[..., 0], expansion)
    gy = _filter_and_adjust_resolution(grad[..., 1], expansion)
    grad = normalize(jnp.stack([gx, gy], axis=-1))

    elementwise_alignment_weight = _field_magnitude(grad)

    # Provide a dummy gradient for the 1D case, which avoids possible nans in the
    # Newton solve below. The tangent vector field will be manually specified.
    is_1d, grad_angle = _is_1d_field(grad)
    dummy_grad = jnp.broadcast_to(jnp.asarray([1, 1], dtype=complex), grad.shape)
    grad = jnp.where(is_1d, dummy_grad, grad)

    # Compute the target field with which the tangent field should be aligned.
    target_field = jnp.stack([grad[..., 1], -grad[..., 0]], axis=-1)
    target_field = normalize_elementwise(target_field)
    if use_jones_direct:
        target_field = normalize_jones(target_field)
        initial_field = target_field
    else:
        initial_field = normalize(jnp.stack([grad[..., 1], -grad[..., 0]], axis=-1))
    fourier_field = fft.fft(initial_field, expansion=expansion, axes=(-3, -2))

    flat_fourier_field = fourier_field.flatten()
    for _ in range(steps):
        _, jac, hessian = field_loss_value_jac_and_hessian(
            flat_fourier_field=flat_fourier_field,
            expansion=expansion,
            primitive_lattice_vectors=primitive_lattice_vectors,
            target_field=target_field,
            elementwise_alignment_loss_weight=elementwise_alignment_weight,
            fourier_loss_weight=fourier_loss_weight,
            smoothness_loss_weight=smoothness_loss_weight,
        )
        flat_fourier_field -= jnp.linalg.solve(hessian, jac.conj())
    fourier_field = flat_fourier_field.reshape((expansion.num_terms, 2))

    field = fft.ifft(fourier_field, expansion=expansion, shape=grid_shape, axis=-2)

    # Manually set the tangent field in the 1d case.
    field_1d = jnp.stack([jnp.sin(grad_angle), jnp.cos(grad_angle)])
    field = jnp.where(is_1d, field_1d, field)
    return normalize(field)


# -------------------------------------------------------------------------------------
# Normalization and transform functions.
# -------------------------------------------------------------------------------------


def normalize_jones(field: jnp.ndarray) -> jnp.ndarray:
    """Generates a Jones vector field following the "Jones" method of [2012 Liu]."""
    assert field.shape[-1] == 2
    field = normalize(field)
    magnitude = _field_magnitude(field)

    magnitude_near_zero = jnp.isclose(magnitude, 0.0)
    magnitude_safe = jnp.where(magnitude_near_zero, 1.0, magnitude)
    tx_norm = jnp.where(
        magnitude_near_zero,
        1 / jnp.sqrt(2),
        field[..., 0, jnp.newaxis] / magnitude_safe,
    )
    ty_norm = jnp.where(
        magnitude_near_zero,
        1 / jnp.sqrt(2),
        field[..., 1, jnp.newaxis] / magnitude_safe,
    )

    phi = jnp.pi / 8 * (1 + jnp.cos(jnp.pi * magnitude))
    theta = _angle(tx_norm + 1j * ty_norm)

    jx = jnp.exp(1j * theta) * (tx_norm * jnp.cos(phi) - ty_norm * 1j * jnp.sin(phi))
    jy = jnp.exp(1j * theta) * (ty_norm * jnp.cos(phi) + tx_norm * 1j * jnp.sin(phi))
    return jnp.concatenate([jx, jy], axis=-1)


def normalize_elementwise(field: jnp.ndarray) -> jnp.ndarray:
    """Normalize the elements of `field` to have magnitude `1` everywhere."""
    magnitude = _field_magnitude(field)
    magnitude_safe = jnp.where(jnp.isclose(magnitude, 0), 1, magnitude)
    return field / magnitude_safe


def normalize(field: jnp.ndarray) -> jnp.ndarray:
    """Normalize `field` so that it has maximum magnitude `1`."""
    max_magnitude = _max_field_magnitude(field)
    max_magnitude_safe = jnp.where(jnp.isclose(max_magnitude, 0), 1, max_magnitude)
    return field / max_magnitude_safe


def _field_magnitude(field: jnp.ndarray) -> jnp.ndarray:
    """Return the magnitude of `field`"""
    magnitude_squared = jnp.sum(jnp.abs(field) ** 2, axis=-1, keepdims=True)
    is_zero = magnitude_squared == 0
    magnitude_squared_safe = jnp.where(is_zero, 1.0, magnitude_squared)
    return jnp.where(is_zero, 0.0, jnp.sqrt(magnitude_squared_safe))


def _max_field_magnitude(field: jnp.ndarray) -> jnp.ndarray:
    """Returns the magnitude of the largest component in `field`."""
    return jnp.amax(_field_magnitude(field), axis=(-3, -2), keepdims=True)


def _angle(x: jnp.ndarray) -> jnp.ndarray:
    """Computes `angle(x)` with special logic for when `x` equals zero."""
    # Avoid taking the angle of an array with any near-zero elements, to avoid
    # `nan` in the gradients.
    is_near_zero = jnp.isclose(x, 0.0)
    x_safe = jnp.where(is_near_zero, (1.0 + 0.0j), x)
    return jnp.angle(x_safe)


def _filter_and_adjust_resolution(
    x: jnp.ndarray,
    expansion: basis.Expansion,
) -> jnp.ndarray:
    """Filter `x` and adjust its resolution for the given `expansion`."""
    y = fft.fft(x, expansion=expansion)
    min_shape = fft.min_array_shape_for_expansion(expansion)
    assert x.ndim == 2
    # Singleton dimensions remain singleton.
    doubled_min_shape = (
        min_shape[0] * (2 if x.shape[0] > 1 else 1),
        min_shape[1] * (2 if x.shape[1] > 1 else 1),
    )
    return fft.ifft(y, expansion=expansion, shape=doubled_min_shape)


def _is_1d_field(field: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Determine whether the field varies in one direction only."""
    ref_field = _field_at_max_magnitude(field)
    assert ref_field.shape == (2,)
    ref_angle = _angle(ref_field[0] + 1j * ref_field[1])
    angle = _angle(field[..., 0] + 1j * field[..., 1])
    magnitude = jnp.squeeze(_field_magnitude(field), axis=-1)
    is_1d = jnp.all(
        jnp.isclose(magnitude, 0.0)
        | jnp.isclose(angle, ref_angle - 2 * jnp.pi, atol=_ATOL_1D_FIELD_ANGLE)
        | jnp.isclose(angle, ref_angle - 1 * jnp.pi, atol=_ATOL_1D_FIELD_ANGLE)
        | jnp.isclose(angle, ref_angle, atol=_ATOL_1D_FIELD_ANGLE)
        | jnp.isclose(angle, ref_angle + 1 * jnp.pi, atol=_ATOL_1D_FIELD_ANGLE)
        | jnp.isclose(angle, ref_angle + 2 * jnp.pi, atol=_ATOL_1D_FIELD_ANGLE)
    )
    # If one of the spatial dimensions is a singleton, the field is automatically 1D.
    assert field.ndim == 3
    is_1d = is_1d | (field.shape[0] == 1) | (field.shape[1] == 1)
    return is_1d, ref_angle


def _field_at_max_magnitude(field: jnp.ndarray) -> jnp.ndarray:
    """Return the field at the location where its magnitude is largest."""
    assert field.ndim == 3
    assert field.shape[-1] == 2
    magnitude = _field_magnitude(field)
    assert magnitude.ndim == 3
    assert magnitude.shape[-1] == 1
    magnitude = magnitude.flatten()
    idx = jnp.argmax(magnitude)
    field = field.reshape((-1, 2))
    return field[idx, :]


# -------------------------------------------------------------------------------------
# Loss function
# -------------------------------------------------------------------------------------


def field_loss_value_jac_and_hessian(
    flat_fourier_field: jnp.ndarray,
    expansion: basis.Expansion,
    primitive_lattice_vectors: basis.LatticeVectors,
    target_field: jnp.ndarray,
    elementwise_alignment_loss_weight: jnp.ndarray,
    fourier_loss_weight: float,
    smoothness_loss_weight: float,
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Compute the value, Jacobian, and Hessian of the field loss."""
    assert flat_fourier_field.ndim == 1

    def fn(flat_fourier_field):
        fourier_field = jnp.reshape(flat_fourier_field, (-1, 2))
        value = _field_loss(
            fourier_field=fourier_field,
            expansion=expansion,
            primitive_lattice_vectors=primitive_lattice_vectors,
            target_field=target_field,
            elementwise_alignment_loss_weight=elementwise_alignment_loss_weight,
            fourier_loss_weight=fourier_loss_weight,
            smoothness_loss_weight=smoothness_loss_weight,
        )
        return value, value

    def jac_fn(flat_fourier_field):
        jac, value = jax.jacrev(fn, has_aux=True)(flat_fourier_field)
        return jac, (value, jac)

    hessian, (value, jac) = jax.jacrev(jac_fn, has_aux=True, holomorphic=True)(
        flat_fourier_field
    )

    return value, jac, hessian


def _field_loss(
    fourier_field: jnp.ndarray,
    expansion: basis.Expansion,
    primitive_lattice_vectors: basis.LatticeVectors,
    target_field: jnp.ndarray,
    elementwise_alignment_loss_weight: jnp.ndarray,
    fourier_loss_weight: float,
    smoothness_loss_weight: float,
) -> jnp.ndarray:
    """Compute loss that favors smooth"""
    shape: Tuple[int, int] = target_field.shape[-3:-1]  # type: ignore[assignment]
    field = fft.ifft(
        y=fourier_field,
        expansion=expansion,
        shape=shape,
        axis=-2,
    )
    loss = _alignment_loss(field, target_field, elementwise_alignment_loss_weight)

    # Avoid calculating the fourier loss and smoothness loss if their weights are zero.
    # On some platforms, including a smoothness loss can signifcantly slow the compile
    # times, and so this optimization increases performance.
    with jax.ensure_compile_time_eval():
        assert jnp.size(fourier_loss_weight) == 1
        assert jnp.size(smoothness_loss_weight) == 1
        use_fourier_loss = fourier_loss_weight > 0
        use_smoothness_loss = smoothness_loss_weight > 0

    if use_fourier_loss:
        loss += fourier_loss_weight * _fourier_loss(
            fourier_field, expansion, primitive_lattice_vectors
        )

    if use_smoothness_loss:
        loss += smoothness_loss_weight * _smoothness_loss(
            field, primitive_lattice_vectors
        )

    return loss


def _alignment_loss(
    field: jnp.ndarray,
    target_field: jnp.ndarray,
    elementwise_alignment_loss_weight: jnp.ndarray,
) -> jnp.ndarray:
    """Compute loss that penalizes differences between `field` and `target_field`."""
    assert elementwise_alignment_loss_weight.ndim == field.ndim
    elementwise_loss = jnp.sum(
        jnp.abs(field - target_field) ** 2, axis=-1, keepdims=True
    )
    return jnp.mean(elementwise_alignment_loss_weight * elementwise_loss)


def _fourier_loss(
    fourier_field: jnp.ndarray,
    expansion: basis.Expansion,
    primitive_lattice_vectors: basis.LatticeVectors,
) -> jnp.ndarray:
    """Compute loss that penalizes high frequency Fourier components.

    The loss is scaled for the size of the unit cell, i.e. two unit cells with
    identical `fourier_field` and scaled `primitive_lattice_vectors` will have
    identical loss.

    Args:
        fourier_field: The Fourier field for which the loss is sought.
        expansion: The Fourier expansion for the field.
        primitive_lattice_vectors: Defines the unit cell coordinates.

    Returns:
        The scalar fourier loss value.
    """
    transverse_wavevectors = basis.transverse_wavevectors(
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
        in_plane_wavevector=jnp.zeros((2,), dtype=fourier_field.real.dtype),
    )

    basis_vectors = jnp.stack(
        [primitive_lattice_vectors.u, primitive_lattice_vectors.v],
        axis=-1,
    )
    area = jnp.abs(jnp.linalg.det(basis_vectors))

    kt = jnp.linalg.norm(transverse_wavevectors, axis=-1) * jnp.sqrt(area)
    return jnp.sum(jnp.abs(fourier_field) ** 2 * kt[..., jnp.newaxis] ** 2)


def _smoothness_loss(
    field: jnp.ndarray, basis_vectors: basis.LatticeVectors
) -> jnp.ndarray:
    """Compute loss associated with smoothness of `field`."""
    grads = _vector_field_forward_difference_gradient(field, basis_vectors)
    return jnp.mean(jnp.abs(jnp.asarray(grads)) ** 2)


# -------------------------------------------------------------------------------------
# Gradient calculation and transformation.
# -------------------------------------------------------------------------------------


def compute_gradient(
    arr: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
) -> jnp.ndarray:
    """Computes the gradient of `arr`.

    The gradient is scaled for the size of the unit cell, i.e. two unit cells with
    identical `arr` and scaled `primitive_lattice_vectors` will have identical
    gradient.

    Args:
        arr: The array for which the gradient is sought.
        primitive_lattice_vectors: Defines the unit cell coordinates.

    Returns:
        The gradient, with shape `arr.shape + (2,)`.
    """
    basis_vectors = jnp.stack(
        [primitive_lattice_vectors.u, primitive_lattice_vectors.v],
        axis=-1,
    )

    area = jnp.abs(jnp.linalg.det(basis_vectors))
    basis_vectors /= jnp.sqrt(area)

    batch_dims = arr.ndim - 2
    pad_width = tuple([(0, 0)] * batch_dims + [(1, 1)] * 2)
    arr_padded = jnp.pad(arr, pad_width, mode="wrap")

    axes = tuple(range(-2, 0))
    partial_grad: List[jnp.ndarray]
    partial_grad = jnp.gradient(arr_padded, axis=axes)  # type: ignore[assignment]
    partial_grad = [_unpad(g, pad_width) for g in partial_grad]
    partial_grad = [g * arr.shape[ax] for g, ax in zip(partial_grad, axes)]
    return _transform_gradient(jnp.stack(partial_grad, axis=-1), basis_vectors)


def _vector_field_forward_difference_gradient(
    field: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Computes the gradient of a vector field by forward difference.

    The returned gradients are,

        grad = (grad_field_x, grad_field_y)

    where

        grad_field_x = stack([dfield_x / dx, dfield_y / dy], axis=-1)

    Args:
        field: The field for which the gradient is sought.
        primitive_lattice_vectors: Defines the unit cell coordinates.

    Returns:
        Tuple containing the gradients, each with the same shape as `field`.
    """
    basis_vectors = jnp.stack(
        [primitive_lattice_vectors.u, primitive_lattice_vectors.v],
        axis=-1,
    )

    area = jnp.abs(jnp.linalg.det(basis_vectors))
    basis_vectors /= jnp.sqrt(area)
    return (
        _scalar_field_forward_difference_gradient(field[..., 0], basis_vectors),
        _scalar_field_forward_difference_gradient(field[..., 1], basis_vectors),
    )


def _scalar_field_forward_difference_gradient(
    field: jnp.ndarray,
    basis_vectors: jnp.ndarray,
) -> jnp.ndarray:
    """Computes the gradient of a scalar field by forward difference.

    Args:
        field: The scalar field for which the forward-difference gradient is sought.
        basis_vectors: The vectors defining the space in which `field` is defined.

    Returns:
        The gradient, with shape `field.shape + (dimensions,)`.
    """
    dimension = basis_vectors.shape[-1]
    assert basis_vectors.shape[-2:] == (dimension, dimension)
    axes = range(field.ndim - dimension, field.ndim)
    diffs = [_periodic_forward_difference(field, axis=ax) for ax in axes]
    partial_grad = jnp.stack(diffs, axis=-1)
    return _transform_gradient(partial_grad, basis_vectors)


def _periodic_forward_difference(field: jnp.ndarray, axis: int) -> jnp.ndarray:
    """Computes the forward difference for `field` along `axis`."""
    diff = jnp.roll(field, shift=-1, axis=axis) - field
    return diff * field.shape[axis]


def _transform_gradient(
    partial_grad: jnp.ndarray,
    basis_vectors: jnp.ndarray,
) -> jnp.ndarray:
    """Compute gradient from partial gradient with arbitrary basis vectors."""
    # https://en.wikipedia.org/wiki/Gradient#General_coordinates
    metric_tensor = _metric_tensor(basis_vectors).astype(partial_grad.dtype)
    inverse_metric_tensor = jnp.linalg.inv(metric_tensor)
    result: jnp.ndarray = partial_grad @ inverse_metric_tensor @ basis_vectors.T
    return result


def _metric_tensor(basis_vectors: jnp.ndarray) -> jnp.ndarray:
    """Compute the metric tensor for a non-Cartesian, non-orthogonal basis."""
    return basis_vectors.T @ basis_vectors


def _unpad(arr: jnp.ndarray, pad_width: Tuple[Tuple[int, int], ...]) -> jnp.ndarray:
    """Undoes a pad operation."""
    slices = [slice(lo, size - hi) for (lo, hi), size in zip(pad_width, arr.shape)]
    return arr[tuple(slices)]


# -----------------------------------------------------------------------------
# Library of available vector field generating schemes.
# -----------------------------------------------------------------------------


JONES_DIRECT: str = "jones_direct"
JONES: str = "jones"
NORMAL: str = "normal"
POL: str = "pol"

JONES_DIRECT_FOURIER: str = "jones_direct_fourier"
JONES_FOURIER: str = "jones_fourier"
NORMAL_FOURIER: str = "normal_fourier"
POL_FOURIER: str = "pol_fourier"

FOURIER_LOSS_WEIGHT: float = 0.2
SMOOTHNESS_LOSS_WEIGHT: float = 2.0

FOURIER_LOSS_WEIGHT_JONES_DIRECT: float = 0.05
SMOOTHNESS_LOSS_WEIGHT_JONES_DIRECT: float = 0.5

VectorFn = Callable[
    [jnp.ndarray, basis.Expansion, basis.LatticeVectors],
    Tuple[jnp.ndarray, jnp.ndarray],
]

VECTOR_FIELD_SCHEMES: Dict[str, VectorFn] = {
    JONES_DIRECT: functools.partial(
        compute_field_jones_direct,
        fourier_loss_weight=0.0,
        smoothness_loss_weight=SMOOTHNESS_LOSS_WEIGHT_JONES_DIRECT,
    ),
    JONES: functools.partial(
        compute_field_jones,
        fourier_loss_weight=0.0,
        smoothness_loss_weight=SMOOTHNESS_LOSS_WEIGHT,
    ),
    NORMAL: functools.partial(
        compute_field_normal,
        fourier_loss_weight=0.0,
        smoothness_loss_weight=SMOOTHNESS_LOSS_WEIGHT,
    ),
    POL: functools.partial(
        compute_field_pol,
        fourier_loss_weight=0.0,
        smoothness_loss_weight=SMOOTHNESS_LOSS_WEIGHT,
    ),
    JONES_DIRECT_FOURIER: functools.partial(
        compute_field_jones_direct,
        fourier_loss_weight=FOURIER_LOSS_WEIGHT_JONES_DIRECT,
        smoothness_loss_weight=0.0,
    ),
    JONES_FOURIER: functools.partial(
        compute_field_jones,
        fourier_loss_weight=FOURIER_LOSS_WEIGHT,
        smoothness_loss_weight=0.0,
    ),
    NORMAL_FOURIER: functools.partial(
        compute_field_normal,
        fourier_loss_weight=FOURIER_LOSS_WEIGHT,
        smoothness_loss_weight=0.0,
    ),
    POL_FOURIER: functools.partial(
        compute_field_pol,
        fourier_loss_weight=FOURIER_LOSS_WEIGHT,
        smoothness_loss_weight=0.0,
    ),
}
